"""Module for estimating Azimuth Synchronous Signal (azss)"""
import numpy as np
from numpy.polynomial import legendre as L
from scipy.interpolate import interp1d
from sotodlib import core, tod_ops
from sotodlib.tod_ops import bin_signal, apodize, filters, pca
import logging

logger = logging.getLogger(__name__)


def bin_by_az(aman, signal=None, az=None, azrange=None, bins=100, flags=None,
              apodize_edges=True, apodize_edges_samps=1600,
              apodize_flags=True, apodize_flags_samps=200):
    """
    Bins a signal by azimuth angle.

    Parameters
    ----------
    aman: TOD
        core.AxisManager
    signal: array-like, optional
        numpy array of signal to be binned. If None, the signal is taken from aman.signal.
    az: array-like, optional
        A 1D numpy array representing the azimuth angles. If not provided, the azimuth angles are taken from aman.boresight.az attribute.
    azrange: array-like, optional
        A list specifying the range of azimuth angles to consider for binning. Defaults to None.
        If None, [min(az), max(az)] will be used for binning.
    bins: int or sequence of scalars
        If bins is an int, it defines the number of equal-width bins in the given azrange (100, by default).
        If bins is a sequence, it defines the bin edges, including the rightmost edge, allowing for non-uniform bin widths.
        If ``bins`` is a sequence, ``bins`` overwrite ``azrange``.
    flags: str or RangesMatrix or Ranges, optional
        Flag indicating whether to exclude flagged samples when binning the signal.
        If provided by a string, `aman.flags.get(flags)` is used for the flags.
        Default is no mask applied.
    apodize_edges : bool, optional
        If True, applies an apodization window to the edges of the signal. Defaults to True.
    apodize_edges_samps : int, optional
        The number of samples over which to apply the edge apodization window. Defaults to 1600.
    apodize_flags : bool, optional
        If True, applies an apodization window based on the flags. Defaults to True.
    apodize_flags_samps : int, optional
        The number of samples over which to apply the flags apodization window. Defaults to 200.

    Returns
    -------
    binning_dict: A dictionary containing the binned signal values.
        * 'bin_edges' : float array of bin edges
        * 'bin_centers': center of each bin of azimuth
        * 'bin_counts': counts of binned samples
        * 'binned_signal': binned signal
        * 'binned_signal_sigma': estimated sigma of binned signal
    """
    if apodize_edges:
        weight_for_signal = apodize.get_apodize_window_for_ends(aman, apodize_samps=apodize_edges_samps)
        if isinstance(flags, str):
            flags = aman.flags.get(flags)
        if (flags is not None) and apodize_flags:
            flags_mask = flags.mask()
            # check the flags dimension
            if flags_mask.ndim == 1:
                flag_is_1d = True
            else:
                all_columns_same = np.all(np.all(flags_mask == flags_mask[0, :], axis=0))
                if all_columns_same:
                    flag_is_1d = True
                    flags_mask = flags_mask[0]
                else:
                    flag_is_1d = False

            if flag_is_1d:
                weight_for_signal = weight_for_signal * apodize.get_apodize_window_from_flags(aman, 
                                                                                              flags=flags,
                                                                                              apodize_samps=apodize_flags_samps)
            else:
                weight_for_signal = weight_for_signal[np.newaxis, :] * apodize.get_apodize_window_from_flags(aman, 
                                                                                                             flags=flags, 
                                                                                                             apodize_samps=apodize_flags_samps)
    else:
        if (flags is not None) and apodize_flags:
            weight_for_signal = apodize.get_apodize_window_from_flags(aman, flags=flags, apodize_samps=apodize_flags_samps)
        else:
            weight_for_signal = None
    binning_dict = bin_signal(aman, bin_by=az, signal=signal,
                              range=azrange, bins=bins, flags=flags, weight_for_signal=weight_for_signal)
    return binning_dict

def fit_azss(az, azss_stats, max_mode, fit_range=None):
    """
    Function for fitting Legendre polynomials to signal binned in azimuth.

    Parameters
    ----------
    az: array-like
        azimuth array from main axis manager
    azss_stats: AxisManager
        Axis manager containing binned signal and azimuth used for fitting.
        Created by ``get_azss`` function.
    max_mode: integer
        Highest order Legendre polynomial to include in the fit.
    fit_range: list
        Azimuth range used to renormalized to the [-1,1] range spanned
        by the Legendre polynomials for fitting. Default is the max-min
        span in the ``binned_az`` array passed in via ``azss_stats``.

    Returns
    -------
    model_sig_tod: array-like
        Model fit for each detector size ndets x n_samps
    """
    bin_width = azss_stats.binned_az[1] - azss_stats.binned_az[0]
    m = ~np.isnan(azss_stats.binned_signal[0])  # masks bins without counts
    if np.count_nonzero(m) < max_mode + 1:
        raise ValueError('Number of valid bins is smaller than mode of Legendre function')

    if fit_range is None:
        az_min = np.min(azss_stats.binned_az[m]) - bin_width / 2
        az_max = np.max(azss_stats.binned_az[m]) + bin_width / 2
    else:
        az_min, az_max = fit_range[0], fit_range[1]

    x_legendre = (2 * az - (az_min+az_max)) / (az_max - az_min)
    x_legendre_bin_centers = (2 * azss_stats.binned_az - (az_min+az_max)) / (az_max - az_min)
    x_legendre_bin_centers = np.where(~m, np.nan, x_legendre_bin_centers)

    coeffs = L.legfit(x_legendre_bin_centers[m], azss_stats.binned_signal[:, m].T, max_mode)
    coeffs = coeffs.T
    binned_model = L.legval(x_legendre_bin_centers, coeffs.T)
    binned_model = np.where(~m, np.nan, binned_model)
    sum_of_squares = np.sum(((azss_stats.binned_signal[:, m] - binned_model[:, m])**2), axis=-1)
    redchi2s = sum_of_squares/azss_stats.uniform_binned_signal_sigma**2 / (len(x_legendre_bin_centers[m]) - max_mode - 1)

    mode_names = np.array([f'legendre{mode}' for mode in range(max_mode + 1)], dtype='<U10')
    modes_axis = core.LabelAxis(name='azss_modes', vals=mode_names)
    azss_stats.add_axis(modes_axis)
    azss_stats.wrap('binned_model', binned_model, [(0, 'dets'), (1, 'bin_az_samps')])
    azss_stats.wrap('x_legendre_bin_centers', x_legendre_bin_centers, [(0, 'bin_az_samps')])
    azss_stats.wrap('coeffs', coeffs, [(0, 'dets'), (1, 'azss_modes')])
    azss_stats.wrap('redchi2s', redchi2s, [(0, 'dets')])

    return L.legval(x_legendre, coeffs.T)


def get_azss(aman, signal='signal', az=None, azrange=None, bins=100, flags=None, scan_flags=None,
             apodize_edges=True, apodize_edges_samps=1600, apodize_flags=True, apodize_flags_samps=200,
             apply_prefilt=True, prefilt_cfg=None, prefilt_detrend='linear',
             method='interpolate', max_mode=None, subtract_in_place=False,
             merge_stats=True, azss_stats_name='azss_stats',
             merge_model=True, azss_model_name='azss_model'):
    """
    Derive azss (Azimuth Synchronous Signal) statistics and model from the given axismanager data.
    **NOTE:** This function does not modify the ``signal`` unless ``subtract_in_place = True``.

    Parameters
    ----------
    aman: TOD
        core.AxisManager
    signal: array-like, optional
        A numpy array representing the signal to be used for azss extraction. If not provided, the signal is taken from aman.signal.
    az: array-like, optional
        A 1D numpy array representing the azimuth angles. If not provided, the azimuth angles are taken from aman.boresight.az.
    azrange: list, optional
        A list specifying the range of azimuth angles to consider for binning. Defaults to [-np.pi, np.pi].
        If None, [min(az), max(az)] will be used for binning.
    bins: int or sequence of scalars
        If bins is an int, it defines the number of equal-width bins in the given azrange (100, by default).
        If bins is a sequence, it defines the bin edges, including the rightmost edge, allowing for non-uniform bin widths.
        If `bins` is a sequence, `bins` overwrite `azrange`.
    flags : str or Rannges or RangesMatrix, optional
        Flag indicating whether to exclude flagged samples when binning the signal.
        Default is no mask applied.
    scan_flags : str or Ranges, optional
        Subtract in the scan/time region specified by flags.
        Typically `flags.left_scan` or `flags.right_scan` will be used.
        If we estimate and subtract azss in left going scan only, then
        `flags` should be a "union of glitch_flags and flags.right_scan (or ~flags.left_scan)", and
        `scan_flags` should be `flags.left_scan`.
    apodize_edges : bool, optional
        If True, applies an apodization window to the edges of the signal. Defaults to True.
    apodize_edges_samps : int, optional
        The number of samples over which to apply the edge apodization window. Defaults to 1600.
    apodize_flags : bool, optional
        If True, applies an apodization window based on the flags. Defaults to True.
    apodize_flags_samps : int, optional
        The number of samples over which to apply the flags apodization window. Defaults to 200.
    apply_prefilt : bool, optional
        If True, applies a pre-filter to the signal before azss extraction. Defaults to True.
    prefilt_cfg : dict, optional
        Configuration for the pre-filter. Defaults to {'type': 'sine2', 'cutoff': 0.005, 'trans_width': 0.005}.
    prefilt_detrend : str, optional
        Method for detrending before filtering. Defaults to 'linear'.
    method: str
        The method to use for azss modeling. Options are 'interpolate' and 'fit'.
        In 'interpolate', binned signal is used directly.
        In 'fit', fitting is applied to the binned signal.
        Defaults to 'interpolate'.
    max_mode: integer, optional
        The number of Legendre modes to use for azss when method is 'fit'. Required when method is 'fit'.
    subtract_in_place: bool
        If True, it subtract the modeled tod from original signal. The aman.signal will be modified.
    merge_stats: boolean, optional
        Boolean flag indicating whether to merge the azss statistics with aman. Defaults to True.
    azss_stats_name: string, optional
        The name to assign to the merged azss statistics. Defaults to 'azss_stats'.
    merge_model: boolean, optional
        Boolean flag indicating whether to merge the azss model with the aman. Defaults to True.
    azss_model_name: string, optional
        The name to assign to the merged azss model. Defaults to 'azss_model'.

    Returns
    -------
    Tuple:
        - azss_stats: core.AxisManager
            - azss statistics including: azumith bin centers, bin counts, binned signal, std of each detector-az bin, std of each detector.
            - If ``method=fit`` then also includes: binned legendre model, legendre bin centers, fit coefficients, reduced chi2.
        - model_sig_tod: numpy.array
            - azss model as a function of time either from fits or interpolation depending on ``method`` argument.
    """
    if prefilt_cfg is None:
        prefilt_cfg = {'type': 'sine2', 'cutoff': 0.005, 'trans_width': 0.005}
    prefilt = filters.get_hpf(prefilt_cfg)

    if signal is None:
        # signal_name variable to be deleted when tod_ops.fourier_filter is updated
        signal_name = 'signal'
        signal = aman[signal_name]
    elif isinstance(signal, str):
        signal_name = signal
        signal = aman[signal_name]
    elif isinstance(signal, np.ndarray):
        raise TypeError("Currently ndarray not supported, need update to tod_ops.fourier_filter module to remove signal_name argument.")
    else:
        raise TypeError("Signal must be None, str, or ndarray")

    if scan_flags is None:
        scan_flags = np.ones(aman.samps.count, dtype=bool)
    elif isinstance(scan_flags, str):
        scan_flags = aman.flags.get(scan_flags).mask()
    else:
        scan_flags = scan_flags.mask()

    if apply_prefilt:
        # This requires signal to be a string.
        signal = np.array(tod_ops.fourier_filter(
                aman, prefilt, detrend=prefilt_detrend, signal_name=signal_name)
                )

    if az is None:
        az = aman.boresight.az

    # do binning
    binning_dict = bin_by_az(aman, signal=signal, az=az, azrange=azrange, bins=bins, flags=flags,
                             apodize_edges=apodize_edges, apodize_edges_samps=apodize_edges_samps,
                             apodize_flags=apodize_flags, apodize_flags_samps=apodize_flags_samps,)
    bin_centers = binning_dict['bin_centers']
    bin_counts = binning_dict['bin_counts']
    binned_signal = binning_dict['binned_signal']
    binned_signal_sigma = binning_dict['binned_signal_sigma']
    uniform_binned_signal_sigma = np.nanmedian(binned_signal_sigma, axis=-1)

    azss_stats = core.AxisManager(aman.dets)
    azss_stats.wrap('binned_az', bin_centers, [(0, core.IndexAxis('bin_az_samps', count=bins))])
    azss_stats.wrap('bin_counts', bin_counts, [(0, 'dets'), (1, 'bin_az_samps')])
    azss_stats.wrap('binned_signal', binned_signal, [(0, 'dets'), (1, 'bin_az_samps')])
    azss_stats.wrap('binned_signal_sigma', binned_signal_sigma, [(0, 'dets'), (1, 'bin_az_samps')])
    azss_stats.wrap('uniform_binned_signal_sigma', uniform_binned_signal_sigma, [(0, 'dets')])

    model_sig_tod = get_azss_model(aman, azss_stats, az, method, max_mode, azrange)

    if merge_stats:
        aman.wrap(azss_stats_name, azss_stats)
    if merge_model:
        aman.wrap(azss_model_name, model_sig_tod, [(0, 'dets'), (1, 'samps')])
    if subtract_in_place:
        aman[signal_name][:, scan_flags] -= model_sig_tod.astype(signal.dtype)[:, scan_flags]
    return azss_stats, model_sig_tod


def get_azss_model(aman, azss_stats, az=None, method='interpolate', max_mode=None, azrange=None):
    """
    Function to return the azss template for subtraction given the azss_stats AxisManager
    """
    if az is None:
        az = aman.boresight.az

    if method == 'fit':
        if not isinstance(max_mode, int):
            raise ValueError('max_mode is not provided as integer')
        model = fit_azss(az=az, azss_stats=azss_stats, max_mode=max_mode, fit_range=azrange)

    if method == 'interpolate':
        # mask az bins that has no data and extrapolate
        mask = ~np.any(np.isnan(azss_stats.binned_signal), axis=0)
        f_template = interp1d(azss_stats.binned_az[mask], azss_stats.binned_signal[:, mask], fill_value='extrapolate')
        model = f_template(az)

    if np.any(~np.isfinite(model)):
        logger.warning('azss model has nan. set zero to nan but this may make glitch')
        model[~np.isfinite(model)] = 0
    return model


def subtract_azss(aman, azss_stats, signal='signal', method='interpolate', max_mode=None, azrange=None,
                  scan_flags=None, subtract_name='azss_remove', in_place=False, remove_template=False):
    """
    Subtract the scan synchronous signal (azss) template from the
    signal in the given axis manager.

    Parameters
    ----------
    aman : AxisManager
        The axis manager containing the signal and the azss template.
    azss_stats: AxisManager
        Contains AxisManager from get_azss.
    signal : str, optional
        The name of the field in the axis manager containing the signal to be processed.
        Defaults to 'signal'.
    method: str
        The method to use for azss modeling. Options are 'interpolate' and 'fit'.
    max_mode: integer, optinal
        The number of Legendre modes to use for azss when method is 'fit'. Required when method is 'fit'.
    azrange:
        Azimuth range used for 'fit' method.
    scan_flags: str or Ranges, optional
        Subtract in the scan/time region specified by flags.
        Typically `flags.left_scan` or `flags.right_scan` will be used.
    subtract_name : str, optional
        The name of the field in the axis manager that will store the azss-subtracted signal.
        Only used if in_place is False. Defaults to 'azss_remove'.
    in_place : bool, optional
        If True, the subtraction is done in place, modifying the original signal in the axis manager.
        If False, the result is stored in a new field specified by subtract_name. Defaults to False.

    Returns
    -------
    None
    """
    if signal is None:
        signal_name = 'signal'
        signal = aman[signal_name]
    elif isinstance(signal, str):
        signal_name = signal
        signal = aman[signal_name]
    elif isinstance(signal, np.ndarray):
        if np.shape(signal) != (aman.dets.count, aman.samps.count):
            raise ValueError("When passing signal as ndarray shape must match (n_dets x n_samps).")
        signal_name = None
    else:
        raise TypeError("Signal must be None, str, or ndarray")

    model = get_azss_model(aman, azss_stats, method=method, max_mode=max_mode, azrange=azrange)

    if scan_flags is None:
        scan_flags = np.ones(aman.samps.count, dtype=bool)
    elif isinstance(scan_flags, str):
        scan_flags = aman.flags.get(scan_flags).mask()
    else:
        scan_flags = scan_flags.mask()

    if in_place:
        if signal_name is None:
            signal[:, scan_flags] -= model.astype(signal.dtype)[:, scan_flags]
        else:
            aman[signal_name][:, scan_flags] -= model.astype(aman[signal_name].dtype)[:, scan_flags]
    else:
        subtracted = signal[:, scan_flags] - model.astype(signal.dtype)[:, scan_flags]
        aman.wrap(subtract_name, subtracted, [(0, 'dets'), (1, 'samps')])


def subtract_azss_template(
    aman,
    signal='signal',
    azss='azss_stats',
    method='interpolate',
    scan_flags=None,
    pca_modes=None,
    subtract=True,
):
    """
    Make azss template model that is "common" to all detectors and subtract
    If pca_modes are specified, make template by pca, otherwise make template by weighted mean.

    Parameters
    ----------
    aman: AxisManager
        The axis manager containing the signal and the azss template.
    signal: str or array-like
        numpy array of signal to be binned. If None, the signal is taken from aman.signal.
    azss: str or azss_stats AxisManager
        azss_stats AxisManager generated by `azss.get_azss`
    method : str
        The method to use for azss modeling. Option is 'interpolate' only now.
    scan_flags: str or flags, optional
        Subtract template model in the time region specified by flags.
        Typically `flags.left_scan` or `flags.right_scan` will be used.
    pca_modes: integer, optinal
        Number of pca modes for making azss template
    subtract: boolean, optional
        If true subtract azss template from signal

    Returns
    -------
    azss_template_model

    """
    if isinstance(signal, str):
        signal = aman.get(signal)
    if isinstance(azss, str):
        azss = aman.get(azss)
    if scan_flags is None:
        scan_flags = np.ones(aman.samps.count, dtype=bool)
    elif isinstance(scan_flags, str):
        scan_flags = aman.flags.get(scan_flags).mask()
    else:
        scan_flags = scan_flags.mask()

    mask = ~np.any(np.isnan(azss.binned_signal), axis=0)

    if pca_modes is not None:
        # make pca model
        template = np.full((azss.dets.count, azss.bin_az_samps.count), np.nan)
        if aman.dets.count < pca_modes:
            raise ValueError(f'The number of pca modes {pca_modes} is '
                             f'larger than the number of detectors {aman.dets.count}.')
        p = pca.get_pca(aman, signal=azss.binned_signal[:, mask])
        R = p.R[:, :pca_modes]
        template[:, mask] = R @ R.T @ azss.binned_signal[:, mask]

    else:
        # make commom mode template
        weighted_azss = np.sum(azss.binned_signal * azss.binned_signal_sigma**-2, axis=0)
        weight = np.sum(azss.binned_signal_sigma**-2, axis=0)
        template = weighted_azss / weight

        # Least square fit the amplitude of azss per each detector
        coefs = np.sum(template[np.newaxis, mask] * azss.binned_signal[:, mask] * azss.binned_signal_sigma[:, mask]**-2, axis=1)
        coefs /= np.sum(template[np.newaxis, mask]**2 * azss.binned_signal_sigma[:, mask]**-2, axis=1)
        template = coefs[:, np.newaxis] * template[np.newaxis, :]

    if method == 'interpolate':
        f_template = interp1d(azss.binned_az[mask], template[:, mask], fill_value='extrapolate')
        model = f_template(aman.boresight.az)
    else:
        raise ValueError(f'{method} is not supported yet')

    if np.any(~np.isfinite(model)):
        logger.warning('azss model has nan. set zero to nan but this may make glitch')
        model[~np.isfinite(model)] = 0

    if subtract:
        signal[:, scan_flags] -= model[:, scan_flags]
    return model



"""FFTs and related operations
"""
from dataclasses import dataclass, field
from functools import lru_cache, partial
from typing_extensions import Callable
from numpy.typing import NDArray
import warnings
import numdifftools as ndt
import numpy as np
import pyfftw
import so3g
import sys
from so3g.proj import Ranges
from scipy.optimize import minimize
from scipy.signal import welch
from scipy.stats import chi2
from sotodlib import core, hwp
from sotodlib.tod_ops import detrend_tod


def _get_num_threads():
    # Guess how many threads we should be using in FFT ops...
    return so3g.useful_info().get("omp_num_threads", 4)


def rfft(
    aman,
    detrend="linear",
    resize="zero_pad",
    window=np.hanning,
    axis_name="samps",
    signal_name="signal",
    delta_t=None,
):
    """Return the real fft of aman.signal_name along the axis axis_name.
        Does not change the data in the axis manager.

    Arguments:

        aman: axis manager

        detrend: Method of detrending to be done before ffting. Can
            be 'linear', 'mean', or None. Note that detrending here can be slow
            for large arrays.

        resize: How to resize the axis to increase fft speed. 'zero_pad'
            will increase to the next 2**N. 'trim' will cut out so the
            factorization of N contains only low primes. None will not
            change the axis length and might be quite slow.

        window: a function that takes N are returns an fft window
            Can be None if no windowing

        axis_name: name of axis you would like to fft along

        signal_name: name of the variable in aman to fft

        delta_t: if none, it will look for 'timestamps' in the axis manager
                and will otherwise assume 1. if not None, it should be the
                sampling rate along the axis you're ffting

    Returns:

        fft: the fft'd data

        freqs: the frequencies it is value at (since resizing is an option)
    """

    axis = getattr(aman, axis_name)

    if len(aman._assignments[signal_name]) == 1:
        n_det = 1
        other_idx = None
    elif len(aman._assignments[signal_name]) == 2:
        checks = np.array(
            [x == axis_name for x in aman._assignments[signal_name]], dtype="bool"
        )
        other_idx = np.where(~checks)[0][0]
        other_axis = getattr(aman, aman._assignments[signal_name][other_idx])
        n_det = other_axis.count
    else:
        raise ValueError("rfft only works for 1D or 2D data streams")

    if detrend is None:
        signal = np.atleast_2d(getattr(aman, signal_name))
    else:
        signal = detrend_tod(
            aman, detrend, axis_name=axis_name, signal_name=signal_name, in_place=True
        )

    if other_idx is not None and other_idx != 0:
        signal = signal.transpose()

    if window is not None:
        signal = signal * window(axis.count)[None, :]

    if resize == "zero_pad":
        k = int(np.ceil(np.log(axis.count) / np.log(2)))
        n = 2**k
    elif resize == "trim":
        n = find_inferior_integer(axis.count)
    elif resize is None:
        n = axis.count
    else:
        raise ValueError('resize must be "zero_pad", "trim", or None')

    rfft = RFFTObj.for_shape(n_det, n, "FFTW_FORWARD")
    if resize == "zero_pad":
        rfft.a[:, : axis.count] = signal
        rfft.a[:, axis.count :] = 0
    elif resize == "trim":
        rfft.a[:] = signal[:, :n]
    else:
        rfft.a[:] = signal[:]

    rfft.t_forward()

    if delta_t is None:
        if "timestamps" in aman:
            delta_t = (aman.timestamps[-1] - aman.timestamps[0]) / axis.count
        else:
            delta_t = 1
    freqs = np.fft.rfftfreq(n, delta_t)

    if other_idx is not None and other_idx != 0:
        return rfft.b.transpose(), freqs

    return rfft.b, freqs


def _t_null(direction):
    raise ValueError(f"No {direction} FFT defined")

@dataclass
class RFFTObj:
    """
    Dataclass to store information needed for rfft.

    Attributes:

        n_det: Number of detectors this object was built for.

        n: Number of samples this object was built for.

        a: Buffer for the real part of the FFT.

        b: Buffer for the complex part of the FFT.

        t_forward: Function for performing the forward FFT.

        t_backward: Function for performing the backward FFT.
    """
    n_det: int
    n: int
    a: NDArray[np.float32]
    b: NDArray[np.complex64]
    t_forward: Callable = field(default=partial(_t_null, "forward"))
    t_backward: Callable = field(default=partial(_t_null, "backward"))

    def validate(self, shape):
        if shape != (self.n_det, self.n):
            raise ValueError("Data is wrong shape for the rfft object")

    @classmethod
    def for_shape(cls, n_det, n, direction="FFTW_FORWARD", **kwargs):
        """Build PyFFTW object for fft-ing
    
        Arguments:
    
            n_det: number of detectors (or just the arr.shape[0] for the
                array you are going to fft)
    
            n: number of samples in timestream
    
            direction: fft direction. Can be FFTW_FORWARD, FFTW_BACKWARD, or BOTH
    
            kwargs: additional arguments to pass to pyfftw.FFTW
    
        Returns:
    
            rfft_obj: An instance of RFFTObj
        """
        fftargs = {"threads": _get_num_threads(), "flags": ["FFTW_ESTIMATE"]}
        fftargs.update(kwargs)
    
        a = pyfftw.empty_aligned((n_det, n), dtype="float32")
        b = pyfftw.empty_aligned((n_det, (n + 2) // 2), dtype="complex64")
        
        t_forward = partial(_t_null, "forward") 
        t_backward = partial(_t_null, "backward") 
        if direction == "FFTW_FORWARD":
            t_forward = pyfftw.FFTW(a, b, direction=direction, **fftargs)
        elif direction == "FFTW_BACKWARD":
            t_backward = pyfftw.FFTW(b, a, direction=direction, **fftargs)
        elif direction == "BOTH":
            t_forward = pyfftw.FFTW(a, b, direction="FFTW_FORWARD", **fftargs)
            t_backward = pyfftw.FFTW(b, a, direction="FFTW_BACKWARD", **fftargs)
        else:
            raise ValueError("direction must be FFTW_FORWARD, FFTW_BACKWARD, or BOTH")
    
        rfft_obj = cls(n_det, n, a, b, t_forward, t_backward)
    
        return rfft_obj 

    @classmethod
    def for_tod(cls, tod, direction="FFTW_FORWARD", **kwargs):
        """
        Wrapper around ``for_shape`` that pulls the shape from a 2d array

        Arguments: 

            tod: 2D array to build rfft object for.
    
            direction: fft direction. Can be FFTW_FORWARD, FFTW_BACKWARD, or BOTH
    
            kwargs: additional arguments to pass to pyfftw.FFTW
    
        Returns:
    
            rfft_obj: An instance of RFFTObj
        """
        if len(tod.shape) != 2:
            raise ValueError("Can only build rfft object for 2D arrays")
        return cls.for_shape(tod.shape[0], tod.shape[1], direction, **kwargs)


@lru_cache
def find_inferior_integer(target, primes=(2, 3, 5, 7, 11, 13)):
    """Find the largest integer less than or equal to target whose prime
    factorization contains only the integers listed in primes.

    """
    p = primes[0]
    n = np.floor(np.log(target) / np.log(p))
    best = p**n
    if len(primes) == 1:
        return int(best)
    while n > 0:
        n -= 1
        base = p**n
        best_friend = getattr(find_inferior_integer, "__wrapped__", find_inferior_integer)(target / base, primes[1:])
        if (best_friend * base) >= best:
            best = best_friend * base
    return int(best)


@lru_cache
def find_superior_integer(target, primes=(2, 3, 5, 7, 11, 13)):
    """Find the smallest integer less than or equal to target whose prime
    factorization contains only the integers listed in primes.

    """
    p = primes[0]
    n = np.ceil(np.log(target) / np.log(p))
    best = p**n
    if len(primes) == 1:
        return int(best)
    while n > 0:
        n -= 1
        base = p**n
        best_friend = getattr(find_superior_integer, "__wrapped__", find_superior_integer)(target / base, primes[1:])
        if (best_friend * base) <= best:
            best = best_friend * base
    return int(best)


def calc_psd(
    aman,
    signal=None,
    timestamps=None,
    max_samples=2**18,
    prefer='center',
    freq_spacing=None,
    merge=False,
    merge_suffix=None,
    overwrite=True,
    subscan=False,
    full_output=False,
    **kwargs
):
    """Calculates the power spectrum density of an input signal using signal.welch().
    Data defaults to aman.signal and times defaults to aman.timestamps.
    By default the nperseg will be set to power of 2 closest to the 1/50th of
    the samples used, this can be overridden by providing nperseg or freq_spacing.

    Arguments:
        aman (AxisManager): with (dets, samps) OR (channels, samps)axes.
        signal (float ndarray): data signal to pass to scipy.signal.welch().
        timestamps (float ndarray): timestamps associated with the data signal.
        max_samples (int): maximum samples along sample axis to send to welch.
        prefer (str): One of ['left', 'right', 'center'], indicating what
            part of the array we would like to send to welch if cuts are
            required.
        freq_spacing (float): The approximate desired frequency spacing of the PSD.
            If None the default nperseg of ~1/50th the signal length is used.
            If an nperseg is explicitly passed then that will be used.
        merge (bool): if True merge results into axismanager.
        merge_suffix (str, optional): Suffix to append to the Pxx field name in aman. Defaults to None (merged as Pxx).
        overwrite (bool): if true will overwrite f, Pxx axes.
        subscan (bool): if True, compute psd on subscans.
        full_output: if True this also outputs nseg, the number of segments used for
            welch, for correcting bias of median white noise estimation by calc_wn.
        **kwargs: keyword args to be passed to signal.welch().

    Returns:
        freqs: array of frequencies corresponding to PSD calculated from welch.
        Pxx: array of PSD values.
        nseg: number of segments used for welch. this is returned if full_output is True.
    """
    if signal is None:
        signal = aman.signal

    if ("noverlap" not in kwargs) or \
            ("noverlap" in kwargs and kwargs["noverlap"] != 0):
        warnings.warn('calc_wn will be biased. noverlap argument of welch '
                      'needs to be 0 to get unbiased median white noise estimate.')
    if not full_output:
        warnings.warn('calc_wn will be biased. full_output argument of calc_psd '
                      'needs to be True to get unbiased median white noise estimate.')

    if subscan:
        if full_output:
            freqs, Pxx, nseg = _calc_psd_subscan(aman, signal=signal,
                                                 freq_spacing=freq_spacing,
                                                 full_output=True,
                                                 **kwargs)
        else:
            freqs, Pxx = _calc_psd_subscan(aman, signal=signal,
                                           freq_spacing=freq_spacing,
                                           **kwargs)
        axis_map_pxx = [(0, "dets"), (1, "nusamps"), (2, "subscans")]
        axis_map_nseg = [(0, "subscans")]
    else:
        if timestamps is None:
            timestamps = aman.timestamps

        n_samps = signal.shape[-1]
        if n_samps <= max_samples:
            start = 0
            stop = n_samps
        else:
            offset = n_samps - max_samples
            if prefer == "left":
                offset = 0
            elif prefer == "center":
                offset //= 2
            elif prefer == "right":
                pass
            else:
                raise ValueError(f"Invalid choice prefer='{prefer}'")
            start = offset
            stop = offset + max_samples
        fs = 1 / np.nanmedian(np.diff(timestamps[start:stop]))
        if "nperseg" not in kwargs:
            if freq_spacing is not None:
                nperseg = int(2 ** (np.around(np.log2(fs / freq_spacing))))
            else:
                nperseg = int(2 ** (np.around(np.log2((stop - start) / 50.0))))
            kwargs["nperseg"] = nperseg

        if kwargs["nperseg"] > max_samples:
            nseg = 1
        else:
            nseg = int(max_samples / kwargs["nperseg"])

        freqs, Pxx = welch(signal[:, start:stop], fs, **kwargs)
        axis_map_pxx = [(0, aman.dets), (1, "nusamps")]
        axis_map_nseg = None

    if merge:
        if 'nusamps' not in aman:
            aman.merge(core.AxisManager(core.OffsetAxis("nusamps", len(freqs))))
            aman.wrap("freqs", freqs, [(0,"nusamps")])
        else:
            if len(freqs) != aman.nusamps.count:
                raise ValueError('New freqs does not match the shape of nusamps\
                                To avoid this, use the same value for nperseg')

        if merge_suffix is None:
            Pxx_name = 'Pxx'
        else:
            Pxx_name = f'Pxx_{merge_suffix}'

        if overwrite:
            if Pxx_name in aman._fields:
                aman.move("Pxx", None)
        aman.wrap(Pxx_name, Pxx, axis_map_pxx)

        if full_output:
            if overwrite and "nseg" in aman._fields:
                aman.move("nseg", None)
            aman.wrap("nseg", nseg, axis_map_nseg)

    if full_output:
        return freqs, Pxx, nseg
    else:
        return freqs, Pxx


def _calc_psd_subscan(aman, signal=None, freq_spacing=None, full_output=False, **kwargs):
    """
    Calculate the power spectrum density of subscans using signal.welch().
    Data defaults to aman.signal. aman.timestamps is used for times.
    aman.subscan_info is used to identify subscans.
    See calc_psd for arguments.
    """
    from .flags import get_subscan_signal
    if signal is None:
        signal = aman.signal

    fs = 1 / np.nanmedian(np.diff(aman.timestamps))
    if "nperseg" not in kwargs:
        if freq_spacing is not None:
            nperseg = int(2 ** (np.around(np.log2(fs / freq_spacing))))
        else:
            duration_samps = np.asarray([np.ptp(x.ranges()) if x.ranges().size > 0 else 0 for x in aman.subscan_info.subscan_flags])
            duration_samps = duration_samps[duration_samps > 0]
            nperseg = int(2 ** (np.around(np.log2(np.median(duration_samps) / 4))))
        kwargs["nperseg"] = nperseg

    Pxx, nseg = [], []
    for iss in range(aman.subscan_info.subscans.count):
        signal_ss = get_subscan_signal(aman, signal, iss)
        axis = -1 if "axis" not in kwargs else kwargs["axis"]
        nsamps = signal_ss.shape[axis]
        if nsamps >= kwargs["nperseg"]:
            freqs, pxx_sub = welch(signal_ss, fs, **kwargs)
            Pxx.append(pxx_sub)
            nseg.append(int(nsamps / kwargs["nperseg"]))
        else:
            Pxx.append(np.full((signal.shape[0], kwargs["nperseg"]//2+1), np.nan)) # Add nans if subscan is too short
            nseg.append(np.nan)
    nseg = np.array(nseg)
    Pxx = np.array(Pxx)
    Pxx = Pxx.transpose(1, 2, 0) # Dets, nusamps, subscans
    if full_output:
        return freqs, Pxx, nseg
    else:
        return freqs, Pxx

def calc_wn(aman, pxx=None, freqs=None, nseg=None, low_f=5, high_f=10):
    """
    Function that calculates the white noise level as a median PSD value between
    two frequencies. Defaults to calculation of white noise between 5 and 10Hz.
    Defaults frequency information to a wrapped "freqs" field in aman.

    Arguments
    ---------
        aman (AxisManager):
            Uses aman.freq as frequency information associated with the PSD, pxx.

        pxx (Float array):
            Psd information to calculate white noise. Defaults to aman.pxx

        freqs (1d Float array):
            frequency information related to the psd. Defaults to aman.freqs

        nseg (Int or 1d Int array):
            number of segmnents used for welch. Defaults to aman.nseg. This is
            necessary for debiasing median white noise estimation. welch PSD with
            non-overlapping n segments follows chi square distribution with
            2 * nseg degrees of freedom. The median of chi square distribution is
            biased from its average.

        low_f (Float):
            low frequency cutoff to calculate median psd value. Defaults to 5Hz

        high_f (float):
            high frequency cutoff to calculate median psd value. Defaults to 10Hz

    Returns
    -------
        wn: Float array of white noise levels for each psd passed into argument.
    """
    if freqs is None:
        freqs = aman.freqs

    if pxx is None:
        pxx = aman.Pxx

    if nseg is None:
        nseg = aman.get('nseg')

    if nseg is None:
        warnings.warn('white noise level estimated by median PSD is biased. '
                      'nseg is necessary to debias. Need to use following '
                      'arguments in calc_psd to get correct nseg. '
                      '`noverlap=0, full_output=True`')
        debias = None
    else:
        debias = 2 * nseg / chi2.ppf(0.5, 2 * nseg)

    fmsk = np.all([freqs >= low_f, freqs <= high_f], axis=0)
    if pxx.ndim == 1:
        wn2 = np.median(pxx[fmsk])
    else:
        wn2 = np.median(pxx[:, fmsk], axis=1)
    if debias is not None:
        if pxx.ndim == 3:
            wn2 *= debias[None, :]
        else:
            wn2 *= debias
    wn = np.sqrt(wn2)
    return wn

def noise_model(f, params, **fixed_param):
    """
    Noise model for power spectrum with white noise, and 1/f noise.
    If any fixed param is handed, that parameter is fixed in the fit.
    'alpha' or 'wn' can be fixed.
    params = [wn, fknee, alpha]
    """
    if 'wn' in fixed_param.keys():
        if len(params)==2:
            wn = fixed_param['wn']
            fknee, alpha = params[0], params[1]
        else:
            raise ValueError('The number of fit parameters are invalid.')
    elif 'alpha' in fixed_param.keys():
        if len(params)==2:
            alpha = fixed_param['alpha']
            wn, fknee = params[0], params[1]
        else:
            raise ValueError('The number of fit parameters are invalid.')
    elif len(fixed_param)==0:
        if len(params)==3:
            wn, fknee, alpha = params[0], params[1], params[2]
        else:
            raise ValueError('The number of fit parameters are invalid.')
    else:
        raise ValueError('"alpha" or "wn" can be a fixed parameter.')
    return wn**2 * (1 + (fknee / f) ** alpha)

def neglnlike(params, x, y, bin_size=1, **fixed_param):
    model = noise_model(x, params, **fixed_param)
    output = np.sum((np.log(model) + y / model)*bin_size)
    if not np.isfinite(output):
        return 1.0e30
    return output

def get_psd_mask(aman, psd_mask=None, f=None,
                mask_hwpss=True, hwp_freq=None, max_hwpss_mode=10, hwpss_width=((-0.4, 0.6), (-0.2, 0.2)),
                mask_peak=False, peak_freq=None, peak_width=(-0.002, +0.002),
                merge=True, overwrite=True
):
    """
    Function to get masks for hwpss or single peak in PSD.

    Arguments
    ---------
        aman : AxisManager
            Axis manager which has 'nusamps' axis.
        psd_mask : numpy.ndarray or Ranges
            Existing psd_mask to be updated. If None, a new mask is created.
        f : nparray
            Frequency of PSD of signal. If None, aman.freqs are used.
        mask_hwpss : bool
            If True, hwpss are masked. Defaults to True.
        hwp_freq : float
            HWP frequency. If None, calculated based on aman.hwp_angle
        max_hwpss_mode : int
            Maximum hwpss mode to subtract. 
        hwpss_width : array-like
            If given in float, 
            nf-hwpss will be masked like (n * hwp_freq - width/2) < f < (n * hwp_freq + width/2).
            If given in array like [1.0, 0.4], 
            1f hwpss is masked like (hwp_freq - 0.5) < f < (hwp_freq + 0.5) and 
            nf hwpss are masked like (n * hwp_freq - 0.2) < f < (n * hwp_freq + 0.2).
            If given in array like [[-0.4, 0.6], [-0.2, 0.3]],
            1f hwpss is masked like (hwp_freq - 0.4) < f < (hwp_freq + 0.6) and
            nf are masked like (n * hwp_freq - 0.2) < f < (n * hwp_freq + 0.3).
        mask_peak : bool
            If True, single peak is masked.
        peak_freq : float
            Center frequency of the mask.
        peak_width : tuple
            Range to mask signal. The default masked range will be 
            (peak_freq - 0.002) < f < (peak_freq + 0.002).
        merge : bool
            if True merge results into axismanager.
        mode: str
            if "replace", existing PSD mask is replaced to new mask.
            If "add", new mask range is added to the existing one.
        overwrite: bool
            if true will overwrite aman.psd_mask.

    Returns
    -------
        psd_mask (nusamps): Ranges array. If merge == True, "psd_mask" is added to the aman.  
    """
    if f is None:
        f = aman.freqs
    if psd_mask is None:
        psd_mask = np.zeros(f.shape, dtype=bool)
    elif isinstance(psd_mask, so3g.RangesInt32):
        psd_mask = psd_mask.mask()

    if mask_hwpss:
        hwp_freq = hwp.get_hwp_freq(aman.timestamps, aman.hwp_solution.hwp_angle)
        psd_mask = psd_mask | get_mask_for_hwpss(f, hwp_freq, max_mode=max_hwpss_mode, width=hwpss_width)
    if mask_peak:
        psd_mask = psd_mask | get_mask_for_single_peak(f, peak_freq, peak_width=peak_width)

    psd_mask = Ranges.from_bitmask(psd_mask)
    if merge:
        if overwrite:
            if "psd_mask" in aman:
                aman.move("psd_mask", None)
        if 'nusamps' not in list(aman._axes.keys()):
            aman.merge(core.AxisManager(core.OffsetAxis("nusamps", len(f))))
        aman.wrap("psd_mask", psd_mask, [(0,"nusamps")])
    return psd_mask

def get_binned_psd(
    aman,
    f=None,
    pxx=None,
    unbinned_mode=3,
    base=1.05,
    merge=False, 
    overwrite=True,
):
    """
    Function that masks hwpss in PSD.

    Arguments
    ---------
        aman : AxisManager
            Axis manager which has samps axis aligned with signal.
        f : nparray
            Frequency of PSD of signal.
        pxx : nparray
            PSD of signal.
        mask : bool
            if True calculate binned psd with mask.
        unbinned_mode : int
            First Fourier modes up to this number are left un-binned.
        base : float (> 1)
            Base of the logspace bins. 
        merge : bool
            if True merge results into axismanager.
        overwrite: bool
            if true will overwrite f, pxx axes.

    Returns
    -------
        f_binned, pxx_binned, bin_size: binned frequency and PSD.
    """
    if f is None:
        f = aman.freqs
    if pxx is None:
        pxx = aman.Pxx
            
    f_bin, bin_size = log_binning(f, unbinned_mode=unbinned_mode, base=base, return_bin_size=True)
    pxx_bin = log_binning(pxx, unbinned_mode=unbinned_mode, base=base, return_bin_size=False)

    if merge:
        aman.merge(core.AxisManager(core.OffsetAxis("nusamps_bin", len(f_bin))))
        if overwrite:
            if "freqs_bin" in aman:
                aman.move("freqs_bin", None)
            if "Pxx_bin" in aman:
                aman.move("Pxx_bin", None)
            if "bin_size" in aman:
                aman.move("bin_size", None)
        aman.wrap("freqs_bin", f_bin, [(0,"nusamps_bin")])
        aman.wrap("bin_size", bin_size, [(0,"nusamps_bin")])
        if pxx_bin.ndim > 2 and 'subscans' in aman and pxx_bin.shape[-1] == aman.subscans.count:
            aman.wrap("Pxx_bin", pxx_bin, [(0, "dets"), (1, "nusamps_bin"), (2, "subscans")])
        else:
            aman.wrap("Pxx_bin", pxx_bin, [(0,"dets"),(1,"nusamps_bin")])

    return f_bin, pxx_bin, bin_size


def fit_noise_model(
    aman,
    signal=None,
    f=None,
    pxx=None,
    psdargs={},
    fknee_est=1,
    wn_est=4E-5,
    alpha_est=3.4,
    merge_fit=False,
    lowf=None,
    f_max=100,
    merge_name="noise_fit_stats",
    merge_psd=True,
    mask=False,
    fixed_param=None,
    binning=False,
    unbinned_mode=3,
    base=1.05,
    freq_spacing=None,
    subscan=False
):
    """
    Fits noise model with white and 1/f noise to the PSD of binned signal.
    This uses a least square method since the distrubition of binned points 
    is close to the standard distribution.

    Args
    ----
    aman : AxisManager
        Axis manager which has samps axis aligned with signal.
    signal : nparray
        Signal sized ndets x nsamps to fit noise model to.
        Default is None which corresponds to aman.signal.
    f : nparray
        Frequency of PSD of signal.
        Default is None which calculates f, pxx from signal.
    pxx : nparray
        PSD sized ndets x len(f) which is fit to with model.
        Default is None which calculates f, pxx from signal.
    psdargs : dict
        Dictionary of optional argument for ``scipy.signal.welch``
    fknee_est : float or nparray
        Initial estimation value of knee frequency. If it is given in array,
        initial values are applied for each detector.
    wn_est : float or nparray
        Initial estimation value of white noise. If it is given in array,
        initial values are applied for each detector.
    alpha_est : float or nparray
        Initial estimation value of alpha. If it is given in array,
        initial values are applied for each detector.
    merge_fit : bool
        Merges fit and fit statistics into input axis manager.
    lowf : float
        Minimum frequency to include in the fitting.
        Default is None which selects lowf as the second index of f.
    f_max : float
        Maximum frequency to include in the fitting. This is particularly
        important for lowpass filtered data such as that post demodulation
        if the data is not downsampled after lowpass filtering.
    merge_name : bool
        If ``merge_fit`` is True then addes into axis manager with merge_name.
    merge_psd : bool
        If ``merge_psd`` is True then adds fres and Pxx to the axis manager.
    mask : bool
        If True, (nusamps,) mask is taken from aman.psd_mask, or calculated on the fly.
        Can also be a 1d (nusamps,) bool array True for good samples to keep.
        Can also be a Ranges in which case it will be inverted before application.

    fixed_param : str
        This accepts 'wn' or 'alpha' or None. If 'wn' ('alpha') is given, 
        white noise level (alpha) is fixed to the wn_est (alpha_est).
    binning : bool
        True to bin the psd before fitting.
        The binning is determined by the 'unbinned_mode' and 'base' params.
    unbinned_mode : int
        First Fourier modes up to this number are left un-binned.
    base : float (> 1)
        Base of the logspace bins.
    freq_spacing : float
        The approximate desired frequency spacing of the PSD. Passed to calc_psd.
    subscan : bool
        If True, fit noise on subscans.
    Returns
    -------
    noise_fit_stats : AxisManager
        If merge_fit is False then axis manager with fit and fit statistics
        is returned otherwise nothing is returned and axis manager is wrapped
        into input aman.
    """

    if signal is None:
        signal = aman.signal

    if f is None or pxx is None:
        psdargs['noverlap'] = psdargs.get('noverlap', 0)
        f, pxx, nseg = calc_psd(
            aman,
            signal=signal,
            timestamps=aman.timestamps,
            freq_spacing=freq_spacing,
            merge=merge_psd,
            subscan=subscan,
            full_output=True,
            **psdargs,
        )
    if np.any(mask):
        if isinstance(mask, np.ndarray):
            pass
        elif isinstance(mask, Ranges):
            mask = ~mask.mask()
        elif mask == True:
            if 'psd_mask' in aman:
                mask = ~aman.psd_mask.mask()
            else: # Calculate on the fly
                mask = ~(get_psd_mask(aman, f=f, merge=False).mask())
        else:
            raise ValueError("mask should be an ndarray or True")
        f = f[mask]
        pxx = pxx[:, mask]

    if subscan:
        fit_noise_model_kwargs = {"fknee_est": fknee_est, "wn_est": wn_est, "alpha_est": alpha_est,
                                  "lowf": lowf, "f_max": f_max, "fixed_param": fixed_param,
                                  "binning": binning, "unbinned_mode": unbinned_mode, "base": base,
                                  "freq_spacing": freq_spacing}
        fitout, covout = _fit_noise_model_subscan(aman, signal,  f, pxx, fit_noise_model_kwargs)
        axis_map_fit = [(0, "dets"), (1, "noise_model_coeffs"), (2, aman.subscans)]
        axis_map_cov = [(0, "dets"), (1, "noise_model_coeffs"), (2, "noise_model_coeffs"), (3, aman.subscans)]
    else:
        eix = np.argmin(np.abs(f - f_max))
        if lowf is None:
            six = 1
        else:
            six = np.argmin(np.abs(f - lowf))
        f = f[six:eix]
        pxx = pxx[:, six:eix]
        bin_size = 1
        # binning
        if binning == True:
            f, pxx, bin_size = get_binned_psd(aman, f=f, pxx=pxx, unbinned_mode=unbinned_mode,
                                              base=base, merge=False)
        fitout = np.zeros((aman.dets.count, 3))
        # This is equal to np.sqrt(np.diag(cov)) when doing curve_fit
        covout = np.zeros((aman.dets.count, 3, 3))
        if isinstance(wn_est, (int, float)):
            wn_est = np.full(aman.dets.count, wn_est)
        elif len(wn_est)!=aman.dets.count:
            print('Size of wn_est must be equal to aman.dets.count or a single value.')
            return
        if isinstance(fknee_est, (int, float)):
            fknee_est = np.full(aman.dets.count, fknee_est)
        elif len(fknee_est)!=aman.dets.count:
            print('Size of fknee_est must be equal to aman.dets.count or a single value.')
            return
        if isinstance(alpha_est, (int, float)):
            alpha_est = np.full(aman.dets.count, alpha_est)
        elif len(alpha_est)!=aman.dets.count:
            print('Size of alpha_est must be equal to aman.dets.count or a single value.')
            return
        if fixed_param == None:
            initial_params = np.array([wn_est, fknee_est, alpha_est])
            bounds= ((sys.float_info.min, None), (sys.float_info.min, None), (0, 10))
        if fixed_param == "wn":
            initial_params = np.array([fknee_est, alpha_est])
            fixed = wn_est
            bounds= ((sys.float_info.min, None), (0, 10))
        if fixed_param == "alpha":
            initial_params = np.array([wn_est, fknee_est])
            fixed = alpha_est
            bounds= ((sys.float_info.min, None), (sys.float_info.min, None))

        for i in range(len(pxx)):
            p = pxx[i]
            p0 = initial_params.T[i]
            _fixed = {}
            if fixed_param != None:
                _fixed = {fixed_param: fixed[i]}
            res = minimize(lambda params: neglnlike(params, f, p, bin_size=bin_size, **_fixed),
                           p0, bounds=bounds, method="Nelder-Mead")
            with warnings.catch_warnings():
                warnings.filterwarnings("error")
                try:
                    Hfun = ndt.Hessian(lambda params: neglnlike(params, f, p, bin_size=bin_size, **_fixed), full_output=True)
                    hessian_ndt, _ = Hfun(res["x"])
                    # Inverse of the hessian is an estimator of the covariance matrix
                    # sqrt of the diagonals gives you the standard errors.
                    covout_i = np.linalg.inv(hessian_ndt)            
                except np.linalg.LinAlgError:
                    print(
                        f"Cannot calculate Hessian for detector {aman.dets.vals[i]} skipping. (LinAlgError)"
                    )
                    covout_i = np.full((len(p0), len(p0)), np.nan)
                except IndexError:
                    print(
                        f"Cannot calculate Hessian for detector {aman.dets.vals[i]} skipping. (IndexError)"
                    )
                    covout_i = np.full((len(p0), len(p0)), np.nan)
                except RuntimeWarning as e:
                    covout_i = np.full((len(p0), len(p0)), np.nan)
                    print(f'RuntimeWarning: {e}\n Hessian failed because results are: {res["x"]}, for det: {aman.dets.vals[i]}')
            fitout_i = res.x
            if fixed_param == "wn":
                covout_i = np.insert(covout_i, 0, 0, axis=0)
                covout_i = np.insert(covout_i, 0, 0, axis=1)
                covout_i[0][0] = np.nan
                fitout_i = np.insert(fitout_i, 0, wn_est[i])
            elif fixed_param == "alpha":
                covout_i = np.insert(covout_i, 2, 0, axis=0)
                covout_i = np.insert(covout_i, 2, 0, axis=1)
                covout_i[2][2] = np.nan
                fitout_i = np.insert(fitout_i, 2, alpha_est[i])
            covout[i] = covout_i
            fitout[i] = fitout_i
        axis_map_fit = [(0, "dets"), (1, "noise_model_coeffs")]
        axis_map_cov = [(0, "dets"), (1, "noise_model_coeffs"), (2, "noise_model_coeffs")]

    noise_model_coeffs = ["white_noise", "fknee", "alpha"]

    noise_fit_stats = core.AxisManager(
        aman.dets,
        core.LabelAxis(
            name="noise_model_coeffs", vals=np.array(noise_model_coeffs, dtype="<U11")
        ),
    )
    noise_fit_stats.wrap("fit", fitout, axis_map_fit)
    noise_fit_stats.wrap("cov", covout, axis_map_cov)

    if merge_fit:
        aman.wrap(merge_name, noise_fit_stats)
    return noise_fit_stats

def get_mask_for_hwpss(freq, hwp_freq, max_mode=10, width=((-0.4, 0.6), (-0.2, 0.2))):
    """
    Function that returns boolean array to mask hwpss in PSD.

    Arguments
    ---------
        freq : nparray
            Frequency of PSD of signal.

        hwp_freq : float
            HWP frequency.

        max_mode : int
            Maximum hwpss mode to subtract. 

        width : float/tuple/list/array
            If given in float, hwpss will be masked like
            (hwp_freq - width/2) < f < (hwp_freq + width/2).
            If given in array like [1.0, 0.4], 1f hwpss is masked like
            (hwp_freq - 0.5) < f < (hwp_freq + 0.5) and Nf hwpss are masked like
            (hwp_freq - 0.2) < f < (hwp_freq + 0.2).
            If given in array like [[-0.4, 0.6], [-0.2, 0.3]],
            1f hwpss is masked like (hwp_freq - 0.4) < f < (hwp_freq + 0.6)
            and Nf are masked like (hwp_freq - 0.2) < f < (hwp_freq + 0.3).
            Usually 1f hwpss distrubites wider than other hwpss.

    Returns
    -------
        mask: Boolean array to mask frequency and power of the given PSD. 
            True in this array stands for the index of hwpss to mask.
    """
    if isinstance(width, (float, int)):
        width_minus = -width/2
        width_plus = width/2
        mask_arrays = []
        for n in range(max_mode):
            mask_arrays.append(get_mask_for_single_peak(freq, hwp_freq*(n+1), peak_width=(width_minus, width_plus)))
    elif isinstance(width, (np.ndarray, list, tuple)):
        width = np.array(width)
        if len(width.shape) == 1:
            # mask for 1f
            width_minus = -width[0]/2
            width_plus = width[0]/2
            mask_arrays = [get_mask_for_single_peak(freq, hwp_freq, peak_width=(width_minus, width_plus))]
            # masks for Nf
            width_minus = -width[1]/2
            width_plus = width[1]/2
            for n in range(max_mode-1):
                mask_arrays.append(get_mask_for_single_peak(freq, hwp_freq*(n+2), peak_width=(width_minus, width_plus)))
        elif len(width.shape) == 2:
            # mask for 1f
            width_minus = width[0][0]
            width_plus = width[0][1]
            mask_arrays = [get_mask_for_single_peak(freq, hwp_freq, peak_width=(width_minus, width_plus))]
            # masks for Nf
            width_minus = width[1][0]
            width_plus = width[1][1]
            for n in range(max_mode-1):
                mask_arrays.append(get_mask_for_single_peak(freq, hwp_freq*(n+2), peak_width=(width_minus, width_plus)))
    mask = np.any(np.array(mask_arrays), axis=0)
    return mask
        
        
def get_mask_for_single_peak(f, peak_freq, peak_width=(-0.002, +0.002)):
    """
    Function that returns boolean array to masks single peak (e.g. scan synchronous signal) in PSD.

    Arguments
    ---------
        f : nparray
            Frequency of PSD of signal.

        peak_freq : float
            Center frequency of the mask.

        peak_width : tuple
            Range to mask signal. The default masked range will be 
            (peak_freq - 0.002) < f < (peak_freq + 0.002).

    Returns
    -------
        mask: Boolean array to mask the given PSD. 
            True in this array stands for the index of the single peak to mask.
    """
    mask = (f > peak_freq + peak_width[0])&(f < peak_freq + peak_width[1])
    return mask

def log_binning(psd, unbinned_mode=3, base=1.05, mask=None,
                return_bin_size=False, drop_nan=False):
    """
    Function to bin PSD or frequency. First several Fourier modes are left un-binned.
    Fourier modes higher than that are averaged into logspace bins.

    Parameters
    ----------
    psd : numpy.ndarray
        PSD (or frequency) to be binned. Can be a 1D or 2D array.
    unbinned_mode : int, optional
        First Fourier modes up to this number are left un-binned. Defaults to 3.
    base : float, optional
        Base of the logspace bins. Must be greater than 1. Defaults to 1.05.
    mask : numpy.ndarray, optional
        Mask for psd. If all values in a bin are masked, the value becomes np.nan.
        Should be a 1D array.
    return_bin_size : bool, optional
        If True, the number of data points in the bins are returned. Defaults to False.
    drop_nan : bool, optional
        If True, drop the indices where psd is NaN. Defaults to False.

    Returns
    -------
    binned_psd : numpy.ndarray
        The binned PSD. If the input is 2D, the output will also be 2D with the same number of rows.
    bin_size : numpy.ndarray, optional
        The number of data points in each bin, only returned if return_bin_size is True.
    """
    if base <= 1:
        raise ValueError("base must be greater than 1")

    is_1d = psd.ndim == 1

    # Ensure psd is at least 2D for consistent processing
    psd = np.atleast_2d(psd)
    num_signals, num_samples = psd.shape[:2]
    
    if mask is not None:
        # Ensure mask is at least 2D and has the same shape as psd
        mask = np.atleast_2d(mask)
        if mask.shape[1] != num_samples:
            raise ValueError("Mask must have the same number of columns as psd")
        mask = np.tile(mask, (num_signals,) + psd.shape[2:] + (1,))
        mask = np.moveaxis(mask, -1, 1)
        psd = np.ma.masked_array(psd, mask=mask)
    
    # Initialize the binned PSD and optionally the bin sizes
    binned_psd = np.zeros((num_signals, unbinned_mode + 1) + psd.shape[2:])
    binned_psd[:, :unbinned_mode + 1] = psd[:, :unbinned_mode + 1]
    bin_size = np.ones((num_signals, unbinned_mode + 1)) if return_bin_size else None

    # Determine the number of bins and their indices
    N = int(np.ceil(np.emath.logn(base, num_samples - unbinned_mode)))
    binning_idx = np.unique(np.logspace(base, N, N, base=base, dtype=int) + unbinned_mode - 1)
    
    # Bin the PSD values for each signal
    new_binned_psd = []
    new_bin_size = []
    for start, end in zip(binning_idx[:-1], binning_idx[1:]):
        bin_mean = np.nanmean(psd[:, start:end], axis=1)
        new_binned_psd.append(bin_mean)
        if return_bin_size:
            new_bin_size.append(end - start)
    
    # Convert lists to numpy arrays and concatenate with initial values
    new_binned_psd = np.array(new_binned_psd)
    new_binned_psd = np.moveaxis(new_binned_psd, 0, 1)# Transpose to match dimensions
    binned_psd = np.hstack([binned_psd, new_binned_psd])
    if return_bin_size:
        new_bin_size = np.array(new_bin_size)
        bin_size = np.hstack([bin_size, np.tile(new_bin_size, (num_signals, 1))])

    if drop_nan:
        valid_indices = ~np.isnan(binned_psd).any(axis=0)
        binned_psd = binned_psd[:, valid_indices]
        if return_bin_size:
            bin_size = bin_size[:, valid_indices]

    if is_1d:
        binned_psd = binned_psd.flatten()
        if return_bin_size:
            bin_size = bin_size.flatten()

    if return_bin_size:
        return binned_psd, bin_size
    return binned_psd


def _fit_noise_model_subscan(
    aman,
    signal,
    f,
    pxx,
    fit_noise_model_kwargs,
):
    """
    Fits noise model with white and 1/f noise to the PSD of signal subscans.
    Args are as for fit_noise_model.
    """
    fitout = np.empty((aman.dets.count, 3, aman.subscan_info.subscans.count))
    covout = np.empty((aman.dets.count, 3, 3, aman.subscan_info.subscans.count))

    per_subscan = {}
    for entry in ["fknee_est", "wn_est", "alpha_est"]:
        if (entry in fit_noise_model_kwargs):
            val = fit_noise_model_kwargs[entry]
            if isinstance(val, np.ndarray) and val.ndim > 1 and val.shape[-1] == aman.subscan_info.subscans.count:
                per_subscan[entry] = fit_noise_model_kwargs.pop(entry)

    kwargs = fit_noise_model_kwargs.copy() if len(per_subscan) > 0 else fit_noise_model_kwargs
    for isub in range(aman.subscan_info.subscans.count):
        if np.all(np.isnan(pxx[...,isub])): # Subscan has been fully cut
            fitout[..., isub] = np.full((aman.dets.count, 3), np.nan)
            covout[..., isub] = np.full((aman.dets.count, 3, 3), np.nan)
        else:
            for entry in list(per_subscan.keys()):
                kwargs[entry] = per_subscan[entry][..., isub]
            noise_model = fit_noise_model(aman, f=f, pxx=pxx[...,isub], merge_fit=False, merge_psd=False, subscan=False, **kwargs)

            fitout[..., isub] = noise_model.fit
            covout[..., isub] = noise_model.cov

    return fitout, covout


def build_hpf_params_dict(
    filter_name,
    noise_fit=None,
    filter_params=None
):
    """
    Build the filter parameter dictionary from a provided
    dictionary or from noise fit results.

    Args
    ----
    filter_name : str
        Name of the filter to build the parameter dict for.
    noise_fit: AxisManager
        AxisManager containing the result of the noise model fit sized nparams x ndets.
    filter_params: dict
        Filter parameters dictionary to complement parameters
        derived from the noise fit (or to be used if noise fit is None).
    Returns
    -------
    filter_params : dict
        Returns a dictionary of the median values of the noise model fit parameters
        if noise_fit is not None, otherwise return the provided filter_params.
    """
    if noise_fit is not None:

        pars_mapping = {
            "high_pass_butter4": {
                "fc": "fknee",
            },
            "counter_1_over_f": {
                "fk": "fknee", 
                "n": "alpha"
            },
            "high_pass_sine2": {
                "cutoff": "fknee",
                "width": None
            }
        }

        if filter_name not in pars_mapping.keys():
            raise NotImplementedError(
                f"{filter_name} params from noise fit is not implemented"
            )
        
        noise_fit_array = noise_fit.fit
        noise_fit_params = noise_fit.noise_model_coeffs.vals
        
        median_params = np.median(noise_fit_array, axis=0)
        median_dict = {
            k: median_params[i]
            for i, k in enumerate(noise_fit_params)
        }

        params_dict = {}
        for k, v in pars_mapping[filter_name].items():
            if v is None:
                if (filter_params is None) or (k not in filter_params):
                    raise ValueError(
                        f"Required parameters {k} not found in config "
                         "and cannot be derived from noise fit."
                    )
                else:
                    params_dict.update({k: filter_params[k]})
            else:
                params_dict[k] = median_dict[v]

        filter_params = params_dict
    
    return filter_params



# Copyright (c) 2024-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

import numpy as np

from toast.utils import Logger
from toast.traits import trait_docs, Int, Unicode
from toast.timing import function_timer
from toast.observation import default_values as defaults
from toast.ops import Operator


@trait_docs
class ActSign(Operator):
    """Apply a sign flip to the responsivity of ACT detector data.

    """

    # Class traits

    API = Int(0, help="Internal interface version for this operator")

    det_data = Unicode(
        defaults.det_data, help="Observation detdata key for the timestream data"
    )

    fp_column = Unicode(
        "det_info:optical_sign", help="Focalplane table column with sign factor"
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        return

    @function_timer
    def _exec(self, data, detectors=None, **kwargs):
        log = Logger.get()

        for ob in data.obs:
            fp_data = {
                x: y for x, y in zip(
                    ob.telescope.focalplane.detector_data["name"],
                    ob.telescope.focalplane.detector_data[self.fp_column]
                )
            }
            for det in ob.local_detectors:
                ob.detdata[self.det_data][det, :] *= fp_data[det]

    def _finalize(self, data, **kwargs):
        return

    def _requires(self):
        req = {
            "detdata": self.det_data,
        }
        return req

    def _provides(self):
        prov = {"meta": list(), "shared": list(), "detdata": list()}
        return prov

    def _accelerators(self):
        return list()



# Copyright (c) 2018-2020 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Simons Observatory core routines.

This module has containers and in-memory structures for data and metadata.

"""
from .context import Context, OBSLOADER_REGISTRY

from .axisman import AxisManager
from .axisman import IndexAxis, OffsetAxis, LabelAxis

from .flagman import FlagManager

from .hardware import Hardware

from . import metadata
from . import util

from .resources import get_local_file, RESOURCE_DEFAULTS



# Copyright (c) 2023-2023 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Simulated detector response to point-like sources.

This includes artificial calibration sources (e.g. drone), solar
system objects, and catalogs of point sources.

"""

import numpy as np
import toast
import toast.ops
from astropy import units as u

from .. import ops as so_ops
from .job import workflow_timer


def setup_simulate_source_signal(operators):
    """Add commandline args and operators for artifical sources.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(so_ops.SimSource(name="sim_source", enabled=False))


@workflow_timer
def simulate_source_signal(job, otherargs, runargs, data):
    """Simulate detector response from artificial sources.

    This uses Az/El detector pointing to observe a terrestrial calibration
    source (e.g. a drone).

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    if job_ops.sim_source.enabled:
        job_ops.sim_source.detector_pointing = job_ops.det_pointing_azel_sim
        if job_ops.sim_source.polarization_fraction != 0:
            job_ops.sim_source.detector_weights = job_ops.weights_azel
        job_ops.sim_source.apply(data)


def setup_simulate_sso_signal(operators):
    """Add commandline args and operators for solar system objects.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(so_ops.SimSSO(name="sim_sso", enabled=False))


@workflow_timer
def simulate_sso_signal(job, otherargs, runargs, data):
    """Simulate detector response from solar system objects.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    if job_ops.sim_sso.enabled:
        job_ops.sim_sso.detector_pointing = job_ops.det_pointing_azel_sim
        job_ops.sim_sso.detector_weights = job_ops.weights_azel
        job_ops.sim_sso.apply(data)


def setup_simulate_catalog_signal(operators):
    """Add commandline args and operators for point source catalogs.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(so_ops.SimCatalog(name="sim_catalog", enabled=False))


@workflow_timer
def simulate_catalog_signal(job, otherargs, runargs, data):
    """Simulate detector response from a point source catalog.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    if job_ops.sim_catalog.enabled:
        job_ops.sim_catalog.detector_pointing = job_ops.det_pointing_radec_sim
        job_ops.sim_catalog.apply(data)



# Copyright (c) 2021 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

"""Check tod_ops routines.

"""

import unittest
import numpy as np
import scipy.signal
from numpy.fft import rfftfreq,irfft
from numpy.testing import assert_allclose

from sotodlib import core, tod_ops, sim_flags
import so3g

from ._helpers import mpi_multi


SAMPLE_FREQ_HZ = 100.

def get_tod(sig_type='trendy', ndets=3, nsamps=1000):
    tod = core.AxisManager(
        core.LabelAxis('dets', ['det%i' % i for i in range(ndets)]),
        core.OffsetAxis('samps', nsamps)
    )
    tod.wrap_new('signal', ('dets', 'samps'), dtype='float32')
    tod.wrap_new('timestamps', ('samps',))[:] = (
        np.arange(tod.samps.count) / SAMPLE_FREQ_HZ)
    if sig_type == 'zero':
        pass
    elif sig_type == 'trendy':
        x = np.linspace(0, 1., tod.samps.count)
        tod.signal[:] = [(i+1) + (i+1)**2 * x for i in range(tod.dets.count)]
    elif sig_type == 'white':
        tod.signal = np.random.normal(size=tod.shape)
    elif sig_type == 'red':
        tod.signal = np.random.normal(size=tod.shape)
        tod.signal[:] = np.cumsum(tod.signal, axis=1)
    else:
        raise RuntimeError(f'sig_type={sig_type}?')
    return tod


def get_glitchy_tod(ts, noise_amp=0, ndets=2, npoly=3, poly_coeffs=None):
    """Returns axis manager to test fill_glitches"""
    fake_signal = np.zeros((ndets, len(ts)))
    input_sig = np.zeros((ndets, len(ts)))
    if poly_coeffs is None:
        poly_coeffs = np.random.uniform(0.5, 1.6, npoly)*1e-1
    poly_sig = np.polyval(poly_coeffs, ts-np.mean(ts))
    for nd in range(ndets):
        input_sig[nd] = poly_sig
        fake_signal[nd] = poly_sig
        noise = np.random.normal(0, noise_amp, size=len(ts))
        fake_signal[nd] += noise

    dets = ['det%i' % i for i in range(ndets)]

    tod_fake = core.AxisManager(core.LabelAxis('dets', vals=dets),
                                core.OffsetAxis('samps', count=len(ts)))
    tod_fake.wrap('timestamps', ts, axis_map=[(0, 'samps')])
    tod_fake.wrap('signal', np.atleast_2d(fake_signal),
                  axis_map=[(0, 'dets'), (1, 'samps')])
    tod_fake.wrap('inputsignal', np.atleast_2d(input_sig),
                  axis_map=[(0, 'dets'), (1, 'samps')])
    flgs = core.AxisManager()
    tod_fake.wrap('flags', flgs)
    params = {'n_glitches': 10, 'sig_n_glitch': 10, 'h_glitch': 10,
              'sig_h_glitch': 2}
    sim_flags.add_random_glitches(tod_fake, params=params, signal='signal',
                                  flag='glitches', overwrite=False)
    return tod_fake


class FactorsTest(unittest.TestCase):
    def test_inf(self):
        f = tod_ops.fft_ops.find_inferior_integer
        self.assertEqual(f(257), 256)
        self.assertEqual(f(28), 28)
        self.assertEqual(f(2**2 * 7**8 + 1), 2**2 * 7**8)

    def test_sup(self):
        f = tod_ops.fft_ops.find_superior_integer
        self.assertEqual(f(255), 256)
        self.assertEqual(f(28), 28)
        self.assertEqual(f(2**2 * 7**8 - 1), 2**2 * 7**8)

class PcaTest(unittest.TestCase):
    """Test the pca module."""
    def test_basic(self):
        tod = get_tod('trendy')
        amps0 = tod.signal.max(axis=1) - tod.signal.min(axis=1)
        modes = tod_ops.pca.get_trends(tod, remove=True)
        amps1 = tod.signal.max(axis=1) - tod.signal.min(axis=1)
        print(f'Amplitudes from {amps0} to {amps1}.')
        self.assertTrue(np.all(amps1 < amps0 * 1e-6))

    def test_pca(self):
        tod = get_tod('white')
        x = tod.timestamps / tod.timestamps[-1]
        comp0 = x**2 - x
        comp1 = x**3 - .2
        comp0, comp1 = [c / np.std(c) for c in [comp0, comp1]]
        tod.signal += comp0 * 100
        tod.signal[1] += comp1 * 50
        mod = tod_ops.pca.get_pca_model(tod, n_modes=2)
        tod_ops.pca.add_model(tod, mod, scale=-1)
        assert (tod.signal.std(axis=1) < 1.).all()

        # With a glitch
        tod = get_tod('white')
        tod.signal += comp0 * 100
        tod.signal[1][10] = 1e6
        mask = np.ones(tod.dets.count, bool)
        mask[1] = False

        # The glitch dominates this PCA, and white noise is not recovered.
        pca = tod_ops.pca.get_pca(tod)
        mod = tod_ops.pca.get_pca_model(tod, pca=pca, n_modes=1)
        sig1 = tod_ops.pca.add_model(tod, mod, scale=-1, signal=tod.signal.copy())
        assert (sig1[mask].std(axis=1) > 10.).any()

        # Excluding det with glitch successfully cleans other dets.
        pca = tod_ops.pca.get_pca(tod, mask=mask)
        mod = tod_ops.pca.get_pca_model(tod, pca=pca, n_modes=1)
        sig2 = tod_ops.pca.add_model(tod, mod, scale=-1, signal=tod.signal.copy())
        assert (sig2[mask].std(axis=1) < 1.).all()

    def test_detrend(self):
        tod = get_tod('trendy')
        tod.wrap('sig1d', tod.signal[0], [(0, 'samps')])
        tod.wrap('sig1e', tod.signal[:,0], [(0, 'dets')])
        tod.wrap('sig3d', tod.signal[None], [(1, 'dets'), (2, 'samps')])
        tod_ops.detrend_tod(tod, signal_name='signal', in_place=False, wrap_name='detrended')
        self.assertTrue( np.all(np.mean(tod.signal,axis=-1) != np.mean(tod.detrended, axis=-1)))

        dx = tod_ops.detrend_tod(tod, signal_name='signal', in_place=False, wrap_name=None)
        self.assertTrue( np.all(np.mean(tod.detrended,axis=-1) == np.mean(dx,axis=-1)))
        
        tod_ops.detrend_tod(tod)
        tod_ops.detrend_tod(tod, signal_name='sig1d')
        tod_ops.detrend_tod(tod, signal_name='sig1e', axis_name='dets')
        tod_ops.detrend_tod(tod, signal_name='sig3d')
        tod_ops.detrend_tod(tod, signal_name='sig3d', axis_name='dets')
        with self.assertRaises(ValueError):
            tod_ops.detrend_tod(tod, signal_name='sig1e')

    def test_detrend_inplace(self):
        tod = get_tod('trendy')
        sig_id = id(tod.signal)

        for method in ["linear", "mean", "median"]:
            tod_ops.detrend_tod(tod, method=method, in_place=True)
            self.assertEqual(sig_id, id(tod.signal))

class GapFillTest(unittest.TestCase):
    def test_basic(self):
        """Test linear fill on simple linear data."""
        tod = get_tod('zero')
        # Latest so3g has RangesMatrix.zeros ... use that next time
        flags = so3g.proj.RangesMatrix([so3g.proj.Ranges(tod.samps.count)
                                        for i in range(tod.dets.count)])
        # For detector 1, identify a short gap; make a linear signal
        # that we know exacly how to gap-fill.
        flags[1].add_interval(10, 20)
        gap_mask = flags[1].mask()
        sig = np.linspace(100., 156., tod.signal.shape[1])
        sentinel = 1.7  # Place-holder for bad samples
        atol = 1e-5  # Note float32 dynamic range!

        # Note _method=None should become 'fast' if accelerated
        # routine is available...
        for order in [1,2]:
            for _method in ['slow', None]:
                # Setup signal.
                tod.signal[1] = sig * ~gap_mask
                tod.signal[1][gap_mask] = sentinel

                # Fill with model, inplace.  (Return origisamples in ex.)
                ex = tod_ops.get_gap_fill(tod, flags=flags, swap=True, _method=_method)
                assert_allclose(sig[gap_mask], tod.signal[1][gap_mask], atol=atol)
                # ... check "extraction" has bad samples
                assert_allclose(ex[1].data, sentinel, atol=atol)

                # Reset signal.
                tod.signal[1] = sig * ~gap_mask
                tod.signal[1][gap_mask] = sentinel

                # Compute fill samples and return in ex, leaving tod unmodified.
                ex = tod_ops.get_gap_fill(tod, flags=flags, order=order,
                                          swap=False, _method=_method)
                assert_allclose(tod.signal[1][gap_mask], sentinel)
                # ... check "extraction" has model values.
                assert_allclose(ex[1].data, sig[gap_mask], atol=atol)
    
    def test_fillglitches(self):
        """Tests fill glitches wrapper function"""
        ts = np.arange(0, 1*60, 1/200)
        aman = get_glitchy_tod(ts, ndets=100)
        # test poly fill
        in_place, up, mg = False, False, None
        glitch_filled = tod_ops.gapfill.fill_glitches(aman, use_pca=up,
                                                      in_place=in_place, wrap=mg)
        self.assertTrue(np.max(np.abs(glitch_filled-aman.inputsignal)) < 1e-3)

        # test pca fill
        in_place, up, mg = False, True, None
        glitch_filled = tod_ops.gapfill.fill_glitches(aman, use_pca=up,
                                                      in_place=in_place, wrap=mg)
        print(np.max(np.abs(glitch_filled-aman.inputsignal)))

        # test wrap new field
        in_place, up, mg = False, False, "gap_filled"
        glitch_filled = tod_ops.gapfill.fill_glitches(aman, use_pca=up,
                                                      in_place=in_place, wrap=mg)
        self.assertTrue('gap_filled' in aman._assignments)

class FilterTest(unittest.TestCase):
    def test_basic(self):
        """Test that fourier filters reduce RMS of white noise."""
        tod = get_tod('white')
        tod.wrap('sig1d', tod.signal[0], [(0, 'samps')])
        sigma0 = tod.signal.std(axis=1)
        f0 = SAMPLE_FREQ_HZ
        fc = f0 / 4

        def wrap_iir(N, wn, fs=f0):
            b, a = scipy.signal.butter(N, wn, fs=fs)
            iir_params = core.AxisManager()
            iir_params.wrap('a', a)
            iir_params.wrap('b', b)
            iir_params.wrap('fscale', 1)
            return iir_params

        # A simple IIR filter
        iir_params = wrap_iir(4, fc)

        # Per-wafer IIR filter params (ok if uniform)
        iir_params_multi = core.AxisManager()
        iir_params_multi.wrap('wafer1', wrap_iir(4, fc))
        iir_params_multi.wrap('wafer2', wrap_iir(4, fc))

        for filt in [
                tod_ops.filters.high_pass_butter4(fc),
                tod_ops.filters.high_pass_sine2(fc),
                tod_ops.filters.low_pass_butter4(fc),
                tod_ops.filters.low_pass_sine2(fc),
                tod_ops.filters.gaussian_filter(fc, f_sigma=f0 / 10),
                tod_ops.filters.gaussian_filter(0, f_sigma=f0 / 10),
                tod_ops.filters.iir_filter(iir_params=iir_params),
                tod_ops.filters.iir_filter(iir_params=iir_params_multi),
                tod_ops.filters.iir_filter(
                    a=iir_params.a, b=iir_params.b, fscale=iir_params.fscale),
                tod_ops.filters.iir_filter(
                    iir_params=dict(iir_params._fields.items())),
                tod_ops.filters.identity_filter(),
        ]:
            f = np.fft.fftfreq(tod.samps.count) * f0
            y = filt(f, tod)
            sig_filt = tod_ops.fourier_filter(tod, filt)
            sigma1 = sig_filt.std(axis=1)
            print(f'Filter takes sigma from {sigma0} to {sigma1}')
            if not isinstance(filt, tod_ops.filters.identity_filter):
                self.assertTrue(np.all(sigma1 < sigma0))

        # Confirm fail if not uniform per-wafer
        iir_params_multi.wrap('wafer3', wrap_iir(6, fc))
        filt = tod_ops.filters.iir_filter(iir_params=iir_params_multi)
        with self.assertRaises(ValueError):
            y = filt(f, tod)

        # Check 1d
        filt = tod_ops.filters.high_pass_butter4(fc)
        sig1f = tod_ops.fourier_filter(tod, filt, signal_name='sig1d',
                                       detrend='linear')
        self.assertEqual(sig1f.shape, tod['sig1d'].shape)


@unittest.skipIf(mpi_multi(), "Running with multiple MPI processes")
class JumpfindTest(unittest.TestCase):
    def test_jumpfinder(self):
        """Test that jumpfinder finds jumps in white noise."""
        np.random.seed(0)
        tod = get_tod('white')
        ptp_orig = np.ptp(tod.signal)
        sig_jumps = tod.signal[0]
        jump_locs = np.array([200, 400, 700])
        sig_jumps[jump_locs[0]:] += 10
        sig_jumps[jump_locs[1]:] -= 13
        sig_jumps[jump_locs[2]:] -= 8

        tod.wrap('sig_jumps', sig_jumps, [(0, 'samps')])

        # Find jumps without filtering
        jumps, _, fixed = tod_ops.jumps.find_jumps(tod, signal=tod.sig_jumps, min_size=5, win_size=23, fix=True)
        jumps_nf = jumps.ranges().flatten()
        
        # Remove double counted jumps and round to remove uncertainty
        jumps_nf = np.unique(np.round(jumps_nf, -2))

        # Check that they agree with the input
        self.assertEqual(len(jump_locs), len(jumps_nf))
        self.assertTrue(np.all(np.abs(jumps_nf - jump_locs) == 0))

        # Check height
        jumps_msk = np.zeros_like(sig_jumps, dtype=bool)
        jumps_msk[jumps_nf] = True
        heights = tod_ops.jumps.estimate_heights(sig_jumps, jumps_msk)
        heights = heights[heights.nonzero()].ravel()
        self.assertTrue(np.all(np.abs(np.array([10, -13, -8]) - np.round(heights)) < 3))

        # Check fixing
        ptp_fix = np.ptp(fixed.ravel()[~jumps.buffer(10).mask().ravel()])
        self.assertTrue(ptp_fix < 1.1*ptp_orig)


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2023-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Noise estimation for mapmaking.
"""

import numpy as np
from astropy import units as u
import toast
import toast.ops

from .. import ops as so_ops
from .job import workflow_timer


def setup_noise_estimation(operators):
    """Add commandline args and operators for estimating detector noise.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(
        toast.ops.NoiseEstim(
            name="noise_estim",
            out_model="noise_estim",
            lagmax=1024,
            nbin_psd=64,
            nsum=1,
            naverage=64,
            enabled=False,
        )
    )
    # Fitting is only performed if noise_estim is enabled
    operators.append(
        toast.ops.FitNoiseModel(
            name="noise_estim_fit",
            out_model="noise_estim_fit",
            enabled=True,
        )
    )


@workflow_timer
def noise_estimation(job, otherargs, runargs, data):
    """Run noise estimation and create a best-fit 1/f model.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    log = toast.utils.Logger.get()
    timer = toast.timing.Timer()
    timer.start()

    # Configured operators for this job
    job_ops = job.operators

    if job_ops.noise_estim.enabled:
        # Estimate noise.
        log.info_rank("  Building noise estimate...", comm=data.comm.comm_world)
        job_ops.noise_estim.apply(data)
        log.info_rank(
            "  Finished noise estimate in",
            comm=data.comm.comm_world,
            timer=timer,
        )

    # Only run noise fitting if estimation is enabled
    if job_ops.noise_estim_fit.enabled:
        if not job_ops.noise_estim.enabled:
            log.info_rank(
                "Noise_estim disabled, nothing to fit",
                comm=data.comm.comm_world,
            )
        else:
            # Create a fit to this noise model
            job_ops.noise_estim_fit.noise_model = job_ops.noise_estim.out_model
            log.info_rank(
                "  Running fit to noise estimate...", comm=data.comm.comm_world
            )
            job_ops.noise_estim_fit.apply(data)
            log.info_rank(
                "  Fit 1/f noise model in",
                comm=data.comm.comm_world,
                timer=timer,
            )


def setup_diff_noise_estimation(operators):
    """Add commandline args and operators for signal difference noise estimation.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(
        toast.ops.SignalDiffNoiseModel(
            name="diff_noise_estim",
            noise_model="diff_noise_estim",
            enabled=False,
        )
    )


@workflow_timer
def diff_noise_estimation(job, otherargs, runargs, data):
    """Estimate a simple white noise model for every detector.

    This uses sample differences to estimate the white noise properties.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    if job_ops.diff_noise_estim.enabled:
        job_ops.diff_noise_estim.apply(data)


def select_mapmaking_noise_model(job, otherargs, runargs, data):
    """Select the noise model for mapmaking.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        (str):  The name of the noise model, or None if a model has
            not yet been created.

    """
    log = toast.utils.Logger.get()

    # Configured operators for this job
    job_ops = job.operators

    noise_model = None
    if (
        hasattr(job_ops, "demodulate")
        and job_ops.demodulate.enabled
        and job_ops.demod_noise_estim.enabled
        and job_ops.demod_noise_estim_fit.enabled
    ):
        # We will use the noise estimate made after demodulation
        log.info_rank("  Using demodulated noise model", comm=data.comm.comm_world)
        noise_model = job_ops.demod_noise_estim_fit.out_model
    elif hasattr(job_ops, "diff_noise_estim") and job_ops.diff_noise_estim.enabled:
        # We have a signal-diff noise estimate
        log.info_rank("  Using signal diff noise model", comm=data.comm.comm_world)
        noise_model = job_ops.diff_noise_estim.noise_model
    elif (
        hasattr(job_ops, "noise_estim")
        and job_ops.noise_estim.enabled
        and job_ops.noise_estim_fit.enabled
    ):
        # We have a noise estimate
        log.info_rank("  Using estimated noise model", comm=data.comm.comm_world)
        noise_model = job_ops.noise_estim_fit.out_model
    else:
        have_noise = True
        for ob in data.obs:
            if "noise_model" not in ob:
                have_noise = False
        if have_noise:
            log.info_rank(
                "  Using noise model from data files", comm=data.comm.comm_world
            )
            noise_model = "noise_model"
        if not have_noise:
            # No noise estimates and no external noise models in the data.
            # Do we have the simulated elevation-weighted nominal model?
            have_noise = True
            for ob in data.obs:
                if "elevation_model" not in ob:
                    have_noise = False
            if have_noise:
                log.info_rank(
                    "  Using nominal synthetic noise model", comm=data.comm.comm_world
                )
                noise_model = "elevation_model"
    return noise_model


def select_raw_noise_model(job, otherargs, runargs, data):
    """Select the noise model prior to demodulation.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        (str):  The name of the noise model, or None if a model has
            not yet been created.

    """
    log = toast.utils.Logger.get()

    # Configured operators for this job
    job_ops = job.operators

    noise_model = None
    if hasattr(job_ops, "diff_noise_estim") and job_ops.diff_noise_estim.enabled:
        # We have a signal-diff noise estimate
        log.info_rank("  Using signal diff noise model", comm=data.comm.comm_world)
        noise_model = job_ops.diff_noise_estim.noise_model
    elif (
        hasattr(job_ops, "noise_estim")
        and job_ops.noise_estim.enabled
        and job_ops.noise_estim_fit.enabled
    ):
        # We have a noise estimate
        log.info_rank("  Using estimated noise model", comm=data.comm.comm_world)
        noise_model = job_ops.noise_estim_fit.out_model
    else:
        have_noise = True
        for ob in data.obs:
            if "noise_model" not in ob:
                have_noise = False
        if have_noise:
            log.info_rank(
                "  Using noise model from data files", comm=data.comm.comm_world
            )
            noise_model = "noise_model"
        if not have_noise:
            # No noise estimates and no external noise models in the data.
            # Do we have the simulated elevation-weighted nominal model?
            have_noise = True
            for ob in data.obs:
                if "elevation_model" not in ob:
                    have_noise = False
            if have_noise:
                log.info_rank(
                    "  Using nominal synthetic noise model", comm=data.comm.comm_world
                )
                noise_model = "elevation_model"
    return noise_model



"""Base Class and PIPELINE register for the preprocessing pipeline scripts."""
import os
import copy
import logging
import numpy as np
from .. import core
from so3g.proj import Ranges, RangesMatrix
from scipy.sparse import csr_array
from matplotlib import pyplot as plt


class _Preprocess(object):
    """The base class for Preprocessing modules which defines the required
    functions and keys required in the configurations.

    Each preprocess module has four overwritable functions that are called by
    the processing scripts in site_pipeline. These four functions are each
    controlled by a specific key in a configuration dictionary passed to the
    module on creation.

    The configuration dictionary has 6 special keys: ``name``, ``process``,
    ``calc``, ``save``, ``select``, and ``plot``. ``name`` is the name used to
    register the module with the PIPELINE registry. The other four keys are matched
    to functions in the module, if the key is not present then that function will 
    be skipped when the preprocessing pipeline is run.

    There are two special AxisManagers expected to be part of the preprocessing
    pipeline. ``aman`` is the "standard" time ordered data AxisManager that is
    loaded via our default styles. ``proc_aman`` is the preprocess AxisManager,
    this is carry the data products that will be saved to whatever Metadata
    Archive is connected to the preprocessing pipeline. 
    """    

    def __init__(self, step_cfgs):
        self.process_cfgs = step_cfgs.get("process")
        self.calc_cfgs = step_cfgs.get("calc")
        self.save_cfgs = step_cfgs.get("save")
        self.select_cfgs = step_cfgs.get("select")
        self.plot_cfgs = step_cfgs.get("plot")
        self.skip_on_sim = step_cfgs.get("skip_on_sim", False)
    def process(self, aman, proc_aman, sim=False):
        """ This function makes changes to the time ordered data AxisManager.
        Ex: calibrating or detrending the timestreams. This function will use
        any configuration information under the ``process`` key of the
        configuration dictionary and is not expected to change or alter
        proc_aman.


        Arguments
        ---------
        aman : AxisManager 
            The time ordered data
        proc_aman : AxisManager 
            Any information generated by previous elements in the preprocessing 
            pipeline.
        sim: Bool
            False by default when analyzing data. Should be True when doing 
            Transfer Function simulations and determining which steps should be run.
        """
        if self.process_cfgs is None:
            return aman, proc_aman
        raise NotImplementedError
        
    def calc_and_save(self, aman, proc_aman):
        """ This function calculates data products of some sort off of the time
        ordered data AxisManager.

        Ex: Calcuating the white noise of the timestream. This function will use
        any configuration information under the ``calc`` key of the
        configuration dictionary and can call the save function to make
        changes to proc_aman.

        Arguments
        ---------
        aman : AxisManager 
            The time ordered data
        proc_aman : AxisManager 
            Any information generated by previous elements in the preprocessing 
            pipeline.
        """
        if self.calc_cfgs is None:
            return aman, proc_aman
        raise NotImplementedError
    
    def save(self, proc_aman, *args):
        """ This function wraps new information into the proc_aman and will use
        any configuration information under the ``save`` key of the
        configuration dictionary.

        Arguments
        ---------
        proc_aman : AxisManager 
            Any information generated by previous elements in the preprocessing 
            pipeline.
        args : any
            Any additional information ``calc_and_save`` needs to send to the
            save function.
        """
        if self.save_cfgs is None:
            return
        raise NotImplementedError

    def select(self, meta, proc_aman=None, in_place=True):
        """ This function runs any desired data selection of the preprocessing
        pipeline results. Assumes the pipeline has already been run and that the
        resulting proc_aman is now saved under the ``preprocess`` key in the
        ``meta`` AxisManager loaded via context.

        Ex: removing detectors with white noise above some limit. This function will use
        any configuration information under the ``select`` key.


        Arguments
        ---------
        meta : AxisManager
            Metadata related to the specific observation
        proc_aman : AxisManager
            Optional. Any information generated by previous elements in the
            preprocessing pipeline.
        in_place : bool
            Optional.  Apply selection and return restricted axis manager if
            True, else return the flag array.

        Returns
        -------
        meta : AxisManager
            Metadata where non-selected detectors have been removed
        """

        if self.select_cfgs is None:
            return meta
        raise NotImplementedError

    def plot(self, aman, proc_aman, filename):
        """ This function creates plots using results from ``calc_and_save``.

        Ex: Plotting det bias flags. This function will use
        any configuration information under the ``plot`` key of the
        configuration dictionary.

        Arguments
        ---------
        aman : AxisManager 
            The time ordered data
        proc_aman : AxisManager 
            Any information generated by previous elements in the preprocessing 
            pipeline.
        filename : str
            Filename should be a concatenation of the global ``plot_dir``
            config with a name with process step number and placeholder {name}
            as shown in Pipeline.run().
        """
        if self.plot_cfgs is None:
            return
        raise NotImplementedError

    @classmethod
    def gen_metric(cls, meta, proc_aman):
        """ Generate a QA metric from the output of this process.

        Arguments
        ---------
        meta : AxisManager
            Metadata related to the specific observation
        proc_aman : AxisManager
            The output of the preprocessing pipeline.

        Returns
        -------
        line : dict
            InfluxDB line entry elements to be fed to
            `site_pipeline.monitor.Monitor.record`
        """
        raise NotImplementedError

    @staticmethod
    def register(process_class):
        """Registers a new modules with the PIPELINE"""
        name = process_class.name

        if Pipeline.PIPELINE.get(name) is None:
            Pipeline.PIPELINE[name] = process_class
        else:
            raise ValueError(
                f"Preprocess Module of name {name} is already Registered"
            )

def _zeros_cls( item ):
    """return a callable zeros class that exactly matches type item for use 
    with wrap_new"""
    if isinstance( item, np.ndarray):
        return lambda shape: np.zeros( shape, dtype = item.dtype)
    elif isinstance( item, RangesMatrix):
        return RangesMatrix.zeros
    elif isinstance( item, Ranges ):
        def temp(shape):
            assert len(shape) == 1
            return Ranges( shape[0] )
        return temp
    elif isinstance( item, csr_array):
        return lambda shape: csr_array(tuple(shape), dtype=item.dtype)
    else:
        raise ValueError(f"Cannot find zero type for {type(item)}")

def _reform_csr_array(arr, oidx, nidx, shape):
    # Support function for _expand, to efficiently embed a csr_array
    # in an expanded frame.  Assumptions are that arr is 2d, with
    # shape (dets, samps).  So oidx and nidx each contain (array of
    # idx, slice).
    row_map = {n: o for n, o in zip(nidx[0], oidx[0])}
    col_shift = oidx[1].start - nidx[1].start

    # For coo array you have:
    #    r, c, v = arr.row, arr.col, arr.data
    # and the equivalent for csr is:
    #    r = np.repeat(np.arange(arr.shape[0]), np.diff(arr.indptr))
    #    c = arr.indices
    #    v = arr.data
    # So the expression for r1, here, is equivalent to
    #    r = np.repeat(np.arange(arr.shape[0]), np.diff(arr.indptr))
    #    r1 = [row_map[_r] for _r in r]
    r1 = np.repeat([row_map[_r] for _r in range(arr.shape[0])],
                   np.diff(arr.indptr))
    c1 = arr.indices + col_shift
    v = arr.data
    return csr_array((v, (r1, c1)), shape=shape, dtype=arr.dtype)

def _ranges_matrix_match( o, n, oidx, nidx):
    """align RangesMatrix n entries to RangesMatrix o"""
    assert len(oidx)==len(nidx)
    if len(oidx) > 2:
        raise NotImplemented
    for i, x in zip( oidx[0], nidx[0]):
        o.ranges[i] = _ranges_match( 
            o.ranges[i], n.ranges[x],
            [oidx[1]], [nidx[1]]
        )
    return o.copy()

def _ranges_match( o, n, oidx, nidx):
    """align Ranges n to Ranges o"""
    assert len(oidx)==len(nidx)
    assert len(oidx)==1
    omsk = o.mask()
    nmsk = n.mask()
    omsk[oidx[0]] = nmsk[nidx[0]]
    return Ranges.from_mask(omsk)

def _intersect(new, out):
    '''Get detector and samples intersection between ``new`` and ``out``.'''
    if 'dets' in new._axes:
        _, fs_dets, ns_dets = out.dets.intersection(
            new.dets,
            return_slices=True
        )
    else:
        fs_dets = range(out.dets.count)
        ns_dets = None
    if 'samps' in new._axes:
        _, fs_samps, ns_samps = out.samps.intersection(
            new.samps,
            return_slices=True
        )
    else:
        fs_samps = slice(None)
        ns_samps = None

    return fs_dets, ns_dets, fs_samps, ns_samps

def _wrap_valid_ranges(new, out, valid_name="valid", wrap_name=None):
    """Wraps in a new Ranges field into ``out`` that tracks the current number
    of detectors and samples that intersect with ``new``.
    """
    fs_dets, _, fs_samps, _ = _intersect(new, out)

    x = Ranges( out.samps.count )
    m = x.mask()
    m[fs_samps] = True
    v = Ranges.from_mask(m)

    valid = RangesMatrix(
        [v if i in fs_dets else x for i in range(out.dets.count)]
    )
    if wrap_name:
        if wrap_name in out:
            out.move(wrap_name, None)
        valid_aman = core.AxisManager(out.dets, out.samps)
        valid_aman.wrap(valid_name, valid, [(0,'dets'),(1,'samps')])
        out.wrap(wrap_name, valid_aman)
    else:
        if valid_name in out:
            out.move(valid_name, None)
        out.wrap(valid_name, valid, [(0,'dets'),(1,'samps')])

def _expand(new, full, wrap_valid=True):
    """new will become a top level axismanager in full once it is matched to
    size"""

    fs_dets, ns_dets, fs_samps, ns_samps = _intersect(new, full)

    out = core.AxisManager()
    for k, v in full._axes.items():
        if k in list(new._axes.keys())+['dets','samps']:
            out._axes[k] = v 

    for a in new._axes:
        if a not in out:
            out.add_axis( new[a] )
    for k, v in new._fields.items():
        if isinstance(v, core.AxisManager):
            out.wrap( k, _expand( v, full) )
        else:
            if np.isscalar(v):
                # Skip expansion for wrapped scalars.
                out.wrap(k, v)
                continue
            out.wrap_new( k, new._assignments[k], cls=_zeros_cls(v))
            oidx=[]; nidx=[]
            for ii, a in enumerate(new._assignments[k]):
                if a == 'dets':
                    oidx.append(fs_dets)
                    nidx.append(ns_dets)
                elif a == 'samps':
                    oidx.append(fs_samps)
                    nidx.append(ns_samps)
                else:
                    if (ii == 0) and isinstance(out[k], RangesMatrix): # Treat like dets
                        # _ranges_matrix_match expects oidx[0] and nidx[0] to be list(inds), not slice.
                        # Unknown axes treated as dets if first entry, else like samps. Added to support (subscans, samps) RangesMatrix.
                        if a in full._axes:
                            _, fs, ns = full[a].intersection(new[a], return_slices=True)
                        else:
                            fs = range(new[a].count)
                            ns = range(new[a].count)
                        oidx.append(fs)
                        nidx.append(ns)
                    else: # Treat like samps
                        oidx.append(slice(None))
                        nidx.append(slice(None))
            oidx = tuple(oidx)
            nidx = tuple(nidx)
            if isinstance(out[k], RangesMatrix):
                assert new._assignments[k][-1] == 'samps'
                out[k] = _ranges_matrix_match( out[k], v, oidx, nidx)
            elif isinstance(out[k], Ranges):
                assert new._assignments[k][0] == 'samps'
                out[k] = _ranges_match( out[k], v, oidx, nidx)
            elif isinstance(out[k], csr_array):
                assert tuple(new._assignments[k]) == ('dets', 'samps')
                out[k] = _reform_csr_array(v, oidx, nidx, out[k].shape)
            else:
                try:
                    out[k][oidx] = v[nidx]
                except TypeError:
                    # Skip expansion for scalar array with no axes.
                    out[k] = v
    if wrap_valid:
        _wrap_valid_ranges(new, out)

    return out

def update_full_aman(proc_aman, full, wrap_valid):
    """Copy new fields from proc_aman[dets,samps] over to 
    full[full-dets,full-samps] after correct re-sizing and indexing.

    Arguments
    ----------
    proc_aman: AxisManager
        A preprocess AxisManager from a pipeline run. The dets,samps axes in 
        proc_aman is assumed to be a subset of the dets,samps axes in full
    full: AxisManager
        A full shape AxisManager that begins the pipeline as the original shape
        of the TOD AxisManager
    """
    for fld in proc_aman._fields:
        if fld not in full._fields:
            assert isinstance(proc_aman[fld], core.AxisManager)
            full.wrap( 
                fld,
                _expand( proc_aman[fld], full, wrap_valid=wrap_valid)
            )

class Pipeline(list):
    """This class is designed to create and run pipelines out of a series of
    different preprocessing modules (classes that inherent from _Preprocess). It
    inherits list object. It also contains the registration of all possible
    preprocess modules in Pipeline.PIPELINE
    """

    PIPELINE = {}

    def __init__(self, modules, plot_dir='./', logger=None, wrap_valid=True):
        """
        Arguments
        ---------
        modules: iterable
            A list or other iterable that contains either instantiated
            _Preprocess instances or the configuration dictionary used to
            instantiate a module
        plot_dir: str
            Directory prefix for preprocess plots
        logger: optional
            logging.logger instance used by the pipeline to send updates
        """
        if logger is None:
            logger = logging.getLogger("pipeline")
        self.logger = logger
        self.plot_dir = plot_dir
        self.wrap_valid = wrap_valid
        super().__init__( [self._check_item(item) for item in copy.deepcopy(modules)])
    
    def _check_item(self, item):
        if isinstance(item, _Preprocess):
            return item
        elif isinstance(item, dict):
            name = item.get("name")
            if name is None:
                raise ValueError(f"Processes made from dictionary must have a 'name' key")
            cls = self.PIPELINE.get(name)
            if cls is None:
                raise ValueError(f"'{name}' not registered as a pipeline element")
            return cls(item)
        else:
            raise ValueError(f"Unknown type created a pipeline element")
    
    # make pipeline have all the list pieces
    def append(self, item):
        super().append( self._check_item(item) )
    def insert(self, index, item):
        super().insert(index, self._check_item(item))
    def extend(self, index, other):
        if isinstance(other, type(self)):
            super().extend(other)
        else:
            super().extend( [self._check_item(item) for item in other])
    def __setitem__(self, index, item):
        super().__setitem__(index, self._check_item(item))
    
    def run(self, aman, proc_aman=None, select=True, sim=False, update_plot=False):
        """
        The main workhorse function for the pipeline class. This function takes
        an AxisManager TOD and successively runs the pipeline of preprocessing
        modules on the AxisManager. The order of operations called by run are::

            for process in pipeline:
                process.process()
                process.calc_and_save()
                    process.save() ## called by process.calc_and_save()
                process.select()

        Arguments
        ---------
        aman: AxisManager
            A TOD object. Generally expected to be raw, unprocessed data. This
            axismanager will be edited in place by the process and select
            functions of each preprocess module
        proc_aman: AxisManager (Optional)
            A preprocess axismanager. If this is provided it is assumed that the
            pipeline has previously been run on this specific TOD and has
            returned this preprocess axismanager. In this case, calls to
            ``process.calc_and_save()`` are skipped as the information is
            expected to be present in this AxisManager.
        select: boolean (Optional)
            if True, the aman detector axis is restricted as described in
            each preprocess module. Most pipelines are developed with 
            select=True. Running select=False may produce unstable behavior
        sim: boolean (Optional)
            if running on sim (``sim=True``), proccesses with the flag
            ``skip_on_sim`` will be skipped.
        update_plot: boolean (Optional)
            if True, re-runs plotting (along with processes and selects)
            given ``proc_aman`` is ``aman.preprocess``. This assumes
            ``process.calc_and_save()`` has been run on this aman before and
            has injested flags and other information into ``proc_aman``.

        Returns
        -------
        proc_aman: AxisManager
            A preprocess axismanager that contains all data products calculated
            throughout the running of the pipeline
        
        """
        if proc_aman is None:
            if 'preprocess' in aman:
                proc_aman = aman.preprocess.copy()
                full = aman.preprocess.copy()
            else:
                proc_aman = core.AxisManager(aman.dets, aman.samps)
                full = core.AxisManager( aman.dets, aman.samps)
            run_calc = True
            update_plot = False
        else:
            if aman.dets.count != proc_aman.dets.count or not np.all(aman.dets.vals == proc_aman.dets.vals):
                self.logger.warning("proc_aman has different detectors than aman. Cutting aman to match")
                det_list = [det for det in proc_aman.dets.vals if det in aman.dets.vals]
                aman.restrict('dets', det_list)
                proc_aman.restrict('dets', det_list)
            full = proc_aman.copy()
            run_calc = False

        if 'frequency_cutoffs' not in proc_aman:
            freq_cutoff = np.nan
            proc_aman.wrap('frequency_cutoffs', core.AxisManager())
            for _field, _sub_iir_params in aman.iir_params._fields.items():
                if isinstance(_sub_iir_params, core.AxisManager):
                    if 'a' in _sub_iir_params._fields:
                        if _sub_iir_params['a'] is not None:
                            from ..tod_ops import filters
                            n = len(aman.timestamps)
                            delta_t = (aman.timestamps[-1] - aman.timestamps[0])/n
                            freqs = np.fft.rfftfreq(n, delta_t)
                            iir = filters.iir_filter()(freqs, aman)

                            mag = np.abs(iir) / np.max(np.abs(iir))
                            # 3dB scale
                            scale = 10 ** (-3. / 20)
                            freq_cutoff = freqs[np.min(np.where(np.array(mag < scale * np.max(mag)))[0])]

            proc_aman['frequency_cutoffs'].wrap('signal', freq_cutoff)

        success = 'end'
        for step, process in enumerate(self):
            if sim and process.skip_on_sim:
                continue
            self.logger.debug(f"Running {process.name}")
            aman, proc_aman = process.process(aman, proc_aman, sim)
            if run_calc:
                aman, proc_aman = process.calc_and_save(aman, proc_aman)
                process.plot(aman, proc_aman, filename=os.path.join(self.plot_dir, '{ctime}/{obsid}', f'{step+1}_{{name}}.png'))
                update_full_aman( proc_aman, full, self.wrap_valid)
            if update_plot:
                process.plot(aman, proc_aman, filename=os.path.join(self.plot_dir, '{ctime}/{obsid}', f'{step+1}_{{name}}.png'))
            plt.close()
            if select:
                process.select(aman, proc_aman)
                proc_aman.restrict('dets', aman.dets.vals)
            self.logger.debug(f"{proc_aman.dets.count} detectors remaining")
            
            if aman.dets.count == 0:
                success = process.name
                break

        if run_calc:
           _wrap_valid_ranges(proc_aman, full, valid_name='valid_data',
                              wrap_name='valid_data')

        # copy updated frequency cutoffs to full
        if "frequency_cutoffs" in full:
            full.move("frequency_cutoffs", None)
        full.wrap("frequency_cutoffs", proc_aman["frequency_cutoffs"])

        return full, success
        

class _FracFlaggedMixIn(object):

    @classmethod
    def gen_metric(
        cls,
        meta,
        proc_aman,
        flags_key=None,
        percentiles=[0, 50, 75, 90, 95, 100],
        tags={},
    ):
        """ Generate a QA metric from the output of this process.

        Arguments
        ---------
        meta : AxisManager
            The full metadata container.
        proc_aman : AxisManager
            The metadata containing just the output of this process.
        flags_key : str or tuple
            The keys to access the flag dataset. e.g. 'glitches.glitch_flags'.
            If a single term is passed, the first item will be inferred to be the
            process name.
        percentiles : list
            Percentiles to compute across detectors
        tags : dict
            The values are keys into `metadata.det_info` to record as tags with
            the Influx line. The keys are addded to the default list
            ["wafer_slot", "tel_tube", "bandpass"] with "bandpass" being taken
            from "wafer.bandpass" or "det_cal.bandpass" if the former isn't found.

        Returns
        -------
        line : dict
            InfluxDB line entry elements to be fed to
            `site_pipeline.monitor.Monitor.record`
        """
        # parse key for flags dataset
        if flags_key is None:
            raise ValueError("The flags_key parameter must be specified in the config.")
        else:
            keys = flags_key.split(".")
            if len(keys) == 1:
                key1, key2 = cls.name, keys[0]
            elif len(keys) == 2:
                key1, key2 = keys
            else:
                raise ValueError(f"Could not parse flags_key {flags_key}")

        # add specified tags
        from ..qa.metrics import _get_tag, _has_tag
        tag_keys = {
            "wafer_slot": "wafer_slot",
            "tel_tube": "tel_tube",
        }

        if _has_tag(meta.det_info, 'wafer.bandpass'):
            bandpasses = meta.det_info.wafer.bandpass
            tag_keys["bandpass"] = "wafer.bandpass"
        else:
            bandpasses = meta.det_info.det_cal.bandpass
            tag_keys["bandpass"] = "det_cal.bandpass"

        tag_keys.update(tags)

        tags = []
        vals = []
        # record one metric per wafer slot, per bandpass
        for bp in np.unique(bandpasses):
            for ws in np.unique(meta.det_info.wafer_slot):
                subset = np.where(
                    (meta.det_info.wafer_slot == ws) & (bandpasses == bp)
                )[0]

                # Compute the number of samples that were flagged
                frac_flagged = np.array([
                    np.dot(r.ranges(), [-1, 1]).sum() for r in proc_aman[key1][key2][subset]
                ], dtype=float)

                if len(frac_flagged) > 0:
                    num_valid = np.array([
                        np.dot(r.ranges(), [-1, 1]).sum() for r in proc_aman[key1].valid[subset]
                    ])
                    with np.errstate(divide="ignore"):
                        frac_flagged *= np.where(num_valid > 0, 1 / num_valid, 0)

                    # record percentiles over detectors and fraction of samples flagged
                    perc = np.percentile(frac_flagged, percentiles)
                    mean = frac_flagged.mean()

                    # get the tags for this wafer (all detectors in this subset share these)
                    tags_base = {
                        k: _get_tag(meta.det_info, i, subset[0]) for k, i in tag_keys.items() if _has_tag(meta.det_info, i)
                    }
                    tags_base["telescope"] = meta.obs_info.telescope

                    # add tags and values to respective lists in order
                    tags_perc = [tags_base.copy() for i in range(perc.size)]
                    for i, t in enumerate(tags_perc):
                        t["det_stat"] = f"percentile_{percentiles[i]}"
                    vals += list(perc)
                    tags += tags_perc
                    tags_mean = tags_base.copy()
                    tags_mean["det_stat"] = "mean"
                    vals.append(mean)
                    tags.append(tags_mean)

        obs_time = [meta.obs_info.timestamp] * len(vals)
        return {
            "field": cls._influx_field,
            "values": vals,
            "timestamps": obs_time,
            "tags": tags,
        }



import numpy as np
import csv
import os


def get_wafer_info(array_name, base_config={}, array_config={}, raw=False,
                   include_no_match=True):
    """Assemble the "wafer info" for a UFM.  This involves reading a few
    csv files and tracing the connections between them.

    Args:
      array_name (str): e.g. 'Mv7'.  Capitalization is important, as
        this is used in filenames.
      base_config (dict): Main config dict.  See notes.
      array_config (dict): Used to override names of specific input
        files (see below).  Normally not necessary; see below.
      raw (bool): If True, return all the data dictionaries and the
        raw data
      include_no_match (bool): If True, include a row with det_id
        "NO_MATCH" to represent resonators with failed association.

    Returns:
       A list where each element contains the data for a single
       readout element (anything that might be associated to a
       resonator; including optical detectors, dark detectors, free
       squids, etc.).

    Notes:
      There are 4 files contributing to the construction of the wafer
      info.  Each may be overridden in the array_config dict; the
      values can be relative to ``base_config['array_info_dir']`` or
      absolute.  See source code for default filenames.

      The base_config dict items used here are:
        - ``array_info_dir``: path to where the wiring files can be
          found.
        - ``bandpass_remap``: a dict specifying how to translate
          bandpass center values provided in the copper_layout into
          values that should be assigned to the dets (and used in
          det_id).  E.g. {90: 220, 150: 280}.

    """
    # Figure out the filenames and read the csv.
    base_dir = base_config.get('array_info_dir', './')
    filenames = [
        os.path.join(base_dir, array_config.get(key, default)) for key, default in [
            ('umux_map_file', 'umux_32_map.csv'),
            ('copper_map_file', 'copper_map_corrected.csv'),
            ('copper_layout_file', 'copper_map_corrected_pixel_num_layout.csv'),
            ('mux_pos_to_band_file', f'{array_name}_mux_pos_num_to_mux_band_num.csv'),
        ]
    ]
    umux_map, copper_map, copper_layout, mux_pos_to_band = \
        map(_read_csv_cols, filenames)

    # Reformat, add some columns.

    ## Sometimes is_north is TRUE/FALSE instead of True/False
    mux_pos_to_band['is_north'][:] = [x.capitalize() for x in mux_pos_to_band['is_north']]

    ## The "pin in copper_map is hidden in SQUID_PIN string.
    copper_map['_pin'] = np.array(
        [k.split('_')[3] for k in copper_map['SQUID_PIN']])

    # Remap the bandpass field, e.g. 150 -> 280 or whatever.
    bandpass_remap = base_config.get('bandpass_remap')
    if bandpass_remap is not None:
        # Do it in two steps to support swapping, shrug.
        masks = []
        for k, v in bandpass_remap.items():
            masks.append((str(v), copper_layout['bandpass'] == str(k)))
        for v, s in masks:
            copper_layout['bandpass'][s] = v

    # The umux_map contains all muxable frequencies from 4-8 GHz.  But
    # SO uses 2 chips, and only the 4-6 GHz band from each.  Therefore
    # reformulate umux_map with a new column 'is_north', and repeat
    # the 4-6 GHz band.
    s = np.array(list(map(int, umux_map['mux_band']))) < 14
    umux_map = {k: np.hstack((v[s], v[s])) for k, v in umux_map.items()}
    umux_map['is_north'] = np.hstack((np.zeros(s.sum(), bool),
                                      np.ones(s.sum(), bool))).astype('str')

    # The four data tables are represented as dicts of arrays of
    # strings.  Strings are ok for now since it's mostly straight
    # table joins on equal values. Re-typing is dealt with in final
    # processing.

    # The umux_map contains all the muxable (is_north, freq) pairs.
    # So loop over entries in that, and find what is attached there.
    idx = []
    data_rows = []
    for umux_i in range(len(umux_map['mux_band'])):
        # Matching the resonator to other tables requires the mux_band
        # (0-27) and bond_pad.  Convert the mux_band to the pair
        # (mod_band, is_highband) to lookup in the mux_pos_to_band
        # table.
        this_umux = _row_from_cols(umux_map, umux_i)
        this_pad = this_umux['bond_pad']  # Note this is a str still.

        # Now get the index into the mux_pos table.
        try:
            mux_map_i = _find_in_cols(mux_pos_to_band,
                                      {'mux_band_num': str(this_umux['mux_band']),
                                       'is_north': this_umux['is_north']})
        except ValueError:
            # This is the only way we should lose resonators.
            continue

        # Use the mux_pos_num to look up detector details.
        this_ptb = _row_from_cols(mux_pos_to_band, mux_map_i)
        this_ptb['array_name'] = array_name  # seems appropriate
        mux_pos_num = this_ptb['mux_pos_num']

        copper_i, layout_i = None, None
        this_copper, this_layout = {}, {}

        try:
            copper_i = _find_in_cols(copper_map,
                                     {'Mux chip position': mux_pos_num,
                                      '_pin': this_pad})
        except:
            pass

        if copper_i is not None:
            layout_i = _find_in_cols(copper_layout,
                                     {'mux_layout_position': mux_pos_num,
                                      'bond_pad': this_pad})
            this_copper = _row_from_cols(copper_map, copper_i)
            this_layout = _row_from_cols(copper_layout, layout_i)

        idx.append((umux_i, mux_map_i, copper_i, layout_i))

        data_rows.append((
            this_umux,
            this_ptb,
            this_copper,
            this_layout,
        ))

    if raw:
        return {
            'filenames': filenames,
            'table_names': ['umux', 'pos_to_band', 'copper', 'layout'],
            'tables': [umux_map, mux_pos_to_band, copper_map, copper_layout],
            'indices': idx,
            'rows': data_rows,
        }

    # Process all rows into final format
    output = []
    for row in data_rows:
        output.append(_process_row(row))

    if include_no_match:
        output.append(_process_row(
            [{'det_id': 'NO_MATCH', 'array_name': array_name},{},{},{}]))

    return output


def _process_row(row):
    # Organize and reformat all the data loaded for a resonator /
    # readout element into a wafer_info entry dict.
    #
    # To get a "NO_MATCH" row pass in
    #   ({'det_id': 'NO_MATCH', 'array_name': ...}, {}, {}, {}).

    umux, pos_to_band, copper, layout = row
    def _get(key, cast=lambda x: x, default=None):
        for src in [umux, copper, layout, pos_to_band]:
            if key in src:
                v = src[key]
                return cast(v)
        return default

    # Construct the output; this sets the ordering of the keys; mark
    # ones we need to post-compute as PLACEHOLDER.
    output = {ko: _get(ki, *args) for ko, ki, *args in [
        ('det_id', 'det_id'),
        ('array', 'array_name'),
        ('bond_pad', 'bond_pad', int, -1),
        ('mux_band', 'mux_band', int, -1),
        ('mux_channel', 'mux_channel', int, -1),
        ('mux_subband', 'mux_subband', str, 'NC'),
        ('mux_position', 'mux_pos_num', int, -1),
        ('design_freq_mhz', 'freq_mhz', float, np.nan),
        ('bias_line', 'Bias line', int, -1),
        ('pol', 'pol', str, 'X'),
        ('bandpass', 'bandpass', str, 'NC'),
        ('det_row', 'DTPixelrow', int, -1),
        ('det_col', 'DTPixelcolumn', int, -1),
        ('rhombus', 'rhomb', str, 'X'),
        ('type', 'PLACEHOLDER'),
        ('x', 'DTPixelxcenter', float, np.nan),
        ('y', 'DTPixelycenter', float, np.nan),
        ('angle', 'DTActualangle', float, np.nan),
        ('crossover', 'DTPadlabel', str, 'X'),
        ('coax', 'is_north', str, None),
    ]}

    if output['bandpass'] != 'NC':
        output['bandpass'] = 'f%03i' % int(output['bandpass'])

    if output['det_id'] == 'NO_MATCH':
        label = 'NO_MATCH'
        det_type = 'NC'
    elif output['det_row'] < 0 or output['pol'] == 'D' or output['bandpass'] == 'NC':
        b, det_type = '%02i' % output['bond_pad'], 'SLOT'
        if b == '64':
            det_type = 'SQID'
        elif b == '-1':
            b, det_type = 'NC', 'BARE'
        elif output['pol'] == 'D':
            det_type = 'DARK'
        elif output['det_row'] == -1:
            det_type = 'UNRT'
        else:
            # SLOT
            output['angle'] = np.nan
            output['crossover'] = 'X'
        mux_pos = '%02i' % _get('mux_pos_num', int)
        label = f'{_get("array_name")}_{det_type}_Mp{mux_pos}b{b}D'
    else:
        det_type = 'OPTC'
        label = '{array}_{bandpass}_{rhombus}r{det_row:02d}c{det_col:02d}{pol}'.format(**output)

    if det_type == 'DARK':
        # The position is taken from 'x' and 'y', not Pixel*center.
        output['x'] = _get('x', float)
        output['y'] = _get('y', float)
        output['angle'] = np.nan

    output.update({
        'det_id': label,
        'type': det_type,
        'coax': {'True': 'N', 'False': 'S'}.get(output['coax'], 'X'),
        'x': output['x'] / 1000.,
        'y': output['y'] / 1000.,
        'array': output['array'].lower(),
        'crossover': output['crossover'][0],
    })

    return output


# CSV -> dict of columns.

def _read_csv_cols(filename):
    rows = []
    with open(filename, newline='') as csvfile:
        reader = csv.reader(csvfile)#, dialect='memberdb')
        for row in reader:
            rows.append(row)
    hdr = rows.pop(0)
    return {h: np.array(c) for h, c in zip(hdr, list(zip(*rows)))}


def _find_in_cols(cols, search_dict):
    idx = [tuple(r) for r in zip(*[cols[k] for k in search_dict])]
    return idx.index(tuple(search_dict.values()))


def _row_from_cols(cols, idx):
    return {k: v[idx] for k, v in cols.items()}
